from functools import reduce
from typing import Union, List

import pandas as pd
from sklearn.pipeline import FeatureUnion, _name_estimators


class DatetimeIndexedFeatureUnion(FeatureUnion):
    """A replacement for FeatureUnion from sklearn.

    Standard sklearn FeatureUnion concatenates features as numpy arrays.
    Here features are expected to be pd.Series or pd.DataFrame,
    and instead of concatenation, they are merged on index.
    """

    def _hstack(self, Xs: List[Union[pd.Series, pd.DataFrame]]) -> pd.DataFrame:
        if not all([isinstance(df.index, pd.DatetimeIndex) for df in Xs]):
            raise ValueError("All features should be in datetime-indexed Series or DataFrames")
        return reduce(lambda left, right: pd.merge(left, right, left_index=True, right_index=True), Xs)


def make_union_of_datetime_indexed_features(*transformers, n_jobs=None, verbose=False):
    """Construct a DatetimeIndexedFeatureUnion from the given transformers.

    Simple reimplementation of `sklearn.pipeline.make_union`
    """
    return DatetimeIndexedFeatureUnion(_name_estimators(transformers), n_jobs=n_jobs, verbose=verbose)



import functools
from typing import Dict, Callable

import numpy as np
import pandas as pd
from statsmodels.tsa.tsatools import add_trend

from hcl_model.model_hcl_generic import HandCraftedLinearModel
from hcl_model.transformers.calendar import CalendarTransformer
from tests.test_model_common import TestModelCommon


class TestHCL(TestModelCommon):
    def test_model_fit(self):
        endog, x_train = self.generate_data()
        y_train = endog[self.lbl_value]

        model = HandCraftedLinearModel()
        model.fit(y=y_train, X=x_train)
        parameters = model.get_parameters()

        params_expected = ["{} const".format(model.lbl_original_exog), "{} time".format(model.lbl_original_exog)]
        assert list(parameters.index) == params_expected
        assert set(model.summary()[model.lbl_params].keys()) == set(params_expected)

    def test_model_prediction(self):
        endog, exog = self.generate_data()
        model = HandCraftedLinearModel()
        num_steps = 10

        y_train = endog.loc[endog.index[:-num_steps], self.lbl_value]
        x_train = exog.loc[endog.index[:-num_steps]]
        x_test = exog.loc[endog.index[-num_steps:]]

        model.fit(y=y_train, X=x_train)
        forecast = model.predict(num_steps=num_steps, X=x_test)

        assert isinstance(forecast, pd.DataFrame)
        assert forecast.shape[0] == num_steps
        assert forecast.columns[0] == self.lbl_value
        assert forecast.index.name == self.lbl_date
        assert isinstance(forecast.index, pd.DatetimeIndex)

    def test_model_simulation(self):
        endog, exog = self.generate_data()
        model = HandCraftedLinearModel()
        num_steps = 10
        num_simulations = 5

        y_train = endog.loc[endog.index[:-num_steps], self.lbl_value]
        x_train = exog.loc[endog.index[:-num_steps]]
        x_test = exog.loc[endog.index[-num_steps:]]

        model.fit(y=y_train, X=x_train)
        simulations = model.simulate(num_steps=num_steps, num_simulations=num_simulations, X=x_test)

        assert isinstance(simulations, pd.DataFrame)
        assert simulations.shape == (num_steps, num_simulations)
        assert simulations.index.name == self.lbl_date
        assert isinstance(simulations.index, pd.DatetimeIndex)

    def test_model_percentiles(self):
        endog, exog = self.generate_data()
        model = HandCraftedLinearModel()
        num_steps = 10
        num_simulations = 5
        quantile_levels = [5, 95]

        y_train = endog.loc[endog.index[:-num_steps], self.lbl_value]
        x_train = exog.loc[endog.index[:-num_steps]]
        x_test = exog.loc[endog.index[-num_steps:]]

        model.fit(y=y_train, X=x_train)
        forecast = model.predict(
            num_steps=num_steps, quantile_levels=quantile_levels, num_simulations=num_simulations, X=x_test
        )

        assert isinstance(forecast, pd.DataFrame)
        assert forecast.shape == (num_steps, len(quantile_levels) + 1)
        assert forecast.columns[0] == self.lbl_value
        assert forecast.index.name == self.lbl_date
        assert isinstance(forecast.index, pd.DatetimeIndex)

    def test_model_summary(self):
        endog, exog = self.generate_data()

        model = HandCraftedLinearModel()
        model.fit(y=endog[self.lbl_value], X=exog)

        assert set(model.summary().index) >= {
            model.lbl_aic,
            model.lbl_r2,
            model.lbl_mape,
            model.lbl_resid_mean,
            model.lbl_resid_std,
            model.lbl_resid_skewness,
            model.lbl_resid_kurtosis,
            model.lbl_params,
        }


class TestHCLTransforms:
    lbl_date = "date"
    lbl_value = "value"

    def generate_input(self) -> pd.DataFrame:
        nobs = 30
        endog = pd.Series(
            np.arange(1, nobs + 1) + np.random.normal(size=nobs, scale=1e-1),
            name=self.lbl_value,
            index=pd.date_range("2019-01-01", periods=nobs, freq="W-FRI", name=self.lbl_date),
        )
        data = endog.to_frame()
        data["x2"] = np.random.normal(size=nobs)
        data["x3"] = np.random.normal(size=nobs)
        return data

    @staticmethod
    def _get_exog_transform() -> Dict[str, Callable]:
        return {"const": lambda x: x + 10, "trend": lambda x: -x.shift(2)}

    @staticmethod
    def _get_endog_transform() -> Dict[str, Callable]:
        return {
            "lag1": lambda y: y.shift(1),
            "local_mean": lambda y: y.shift(1).ewm(span=5).mean(),
        }

    def test_transform_lags(self):
        lags = [1, 2, 10]
        col_name = "lag_{}"

        f = {col_name.format(lag): functools.partial(lambda lag, y: y.shift(lag), lag) for lag in lags}
        model = HandCraftedLinearModel(endog_transform=f)
        endog = pd.Series(np.arange(5))

        transformed = model._transform_data(data=endog, transform=f)
        transformed_expected = {col_name.format(lag): endog.shift(lag) for lag in lags}

        for key, val in transformed.items():
            pd.testing.assert_series_equal(val, transformed_expected[key])

    def test_transform_data(self):
        data = self.generate_input()
        f = self._get_endog_transform()
        g = self._get_exog_transform()

        model = HandCraftedLinearModel(endog_transform=f, exog_transform=g)
        endog = data[self.lbl_value]
        exog = data.iloc[:, 1:]

        transformed = model._transform_data(data=endog, transform=f)
        transformed_expected = {key: endog.transform(f[key]) for key in f.keys()}

        for key, val in transformed.items():
            pd.testing.assert_series_equal(val, transformed_expected[key])

        transformed = model._transform_data(data=exog, transform=g)
        transformed_expected = {key: exog.transform(g[key]) for key in g.keys()}

        for key, val in transformed.items():
            pd.testing.assert_frame_equal(val, transformed_expected[key])

        transformed = model._transform_all_data(exog=exog, endog=endog)
        transformed_endog_expected = {key: endog.transform(f[key]) for key in f.keys()}
        transformed_exog_expected = {key: exog.transform(g[key]) for key in g.keys()}

        for key, val in transformed_endog_expected.items():
            pd.testing.assert_series_equal(val, transformed[key])
        for key, val in transformed_exog_expected.items():
            pd.testing.assert_frame_equal(val, transformed[key])

    def test_convert_transformed_dict_to_frame(self):
        data = self.generate_input()
        f = self._get_endog_transform()
        g = self._get_exog_transform()

        model = HandCraftedLinearModel(endog_transform=f, exog_transform=g)
        endog = data[self.lbl_value]
        exog = data.iloc[:, 1:]

        transformed = model._transform_all_data(exog=exog, endog=endog)
        transformed_df = model._convert_transformed_dict_to_frame(transformed=transformed)

        keys = set(f.keys())
        for key in g.keys():
            keys.update({"{} {}".format(key, col) for col in exog.columns})

        assert set(transformed_df.columns) == keys

    def test_model_fit(self):
        data = self.generate_input()
        f = self._get_endog_transform()
        g = self._get_exog_transform()
        y_train = data[self.lbl_value]
        x_train = data.iloc[:, 1:]

        model = HandCraftedLinearModel(endog_transform=f, exog_transform=g)
        model.fit(y=y_train, X=x_train)

        parameters = model.get_parameters()

        keys = set(f.keys())
        for key in g.keys():
            keys.update({"{} {}".format(key, col) for col in x_train.columns})

        # Some random test. No good logic here
        assert set(parameters.index) == keys
        assert parameters.isna().sum() == 0
        assert set(model.summary()[model.lbl_params].keys()) == keys

    def test_model_prediction(self):
        data = self.generate_input()
        f = self._get_endog_transform()
        g = self._get_exog_transform()
        num_steps = 5
        y_train = data.loc[data.index[:-num_steps], self.lbl_value]
        x_train = data.iloc[:-num_steps, 1:]
        x_test = data.iloc[-num_steps:, 1:]

        model = HandCraftedLinearModel(endog_transform=f, exog_transform=g)
        model.fit(y=y_train, X=x_train)

        forecast = model.predict(num_steps=num_steps, X=x_test)

        assert isinstance(forecast, pd.DataFrame)
        assert forecast.shape[0] == num_steps
        assert forecast.columns[0] == self.lbl_value
        assert forecast.isna().sum().sum() == 0
        assert forecast.index.name == self.lbl_date
        assert isinstance(forecast.index, pd.DatetimeIndex)

    def test_model_percentiles(self):
        data = self.generate_input()
        f = self._get_endog_transform()
        g = self._get_exog_transform()

        num_steps = 5
        num_simulations = 5
        quantile_levels = [5, 95]

        model = HandCraftedLinearModel(endog_transform=f, exog_transform=g)
        y_train = data.loc[data.index[:-num_steps], self.lbl_value]
        x_train = data.iloc[:-num_steps, 1:]
        x_test = data.iloc[-num_steps:, 1:]

        model.fit(y=y_train, X=x_train)

        forecast = model.predict(
            num_steps=num_steps, X=x_test, quantile_levels=quantile_levels, num_simulations=num_simulations
        )

        assert isinstance(forecast, pd.DataFrame)
        assert forecast.shape == (num_steps, len(quantile_levels) + 1)
        assert forecast.columns[0] == self.lbl_value
        assert forecast.isna().sum().sum() == 0
        assert forecast.index.name == self.lbl_date
        assert isinstance(forecast.index, pd.DatetimeIndex)

    def test_with_calendar_transform(self):
        data = self.generate_input()
        f = self._get_endog_transform()
        degrees_of_freedom = 4
        lbl_splines = "splines"
        lbl_squared = "squared"
        g = {
            lbl_splines: lambda df: CalendarTransformer().add_periodic_splines(
                df=df, degrees_of_freedom=degrees_of_freedom
            ),
            lbl_squared: lambda df: df ** 2,
        }
        num_steps = 5
        x_train = data.iloc[:-num_steps, 1:]
        y_train = data.loc[data.index[:-num_steps], self.lbl_value]
        x_test = data.iloc[-num_steps:, 1:]

        model = HandCraftedLinearModel(endog_transform=f, exog_transform=g)
        transformed = model._transform_data(data=x_train, transform=g)
        transformed_df = model._convert_transformed_dict_to_frame(transformed=transformed)
        model.fit(y=y_train, X=x_train)

        forecast = model.predict(num_steps=num_steps, X=x_test)

        assert isinstance(forecast, pd.DataFrame)
        assert forecast.shape[0] == num_steps
        assert forecast.columns[0] == self.lbl_value
        assert forecast.isna().sum().sum() == 0
        assert forecast.index.name == self.lbl_date
        assert isinstance(forecast.index, pd.DatetimeIndex)
        assert {lbl_splines, lbl_squared} == set(transformed.keys())
        assert len(transformed[lbl_splines].columns) == x_train.shape[1] + degrees_of_freedom
        assert len(transformed[lbl_squared].columns) == x_train.shape[1]
        assert len(transformed_df.columns) == 2 * x_train.shape[1] + degrees_of_freedom


class TestHCLWeightedTransforms:
    lbl_date = "date"
    lbl_value = "value"

    def generate_input(self):
        nobs = 30
        endog = pd.Series(
            np.arange(1, nobs + 1) + np.random.normal(size=nobs, scale=1e-1),
            name=self.lbl_value,
            index=pd.date_range("2019-01-01", periods=nobs, freq="W-FRI", name=self.lbl_date),
        )

        data = add_trend(endog, trend="ct")
        data["x3"] = 999

        f = {
            "lag1": lambda y: y.shift(1),
            "local_mean": lambda y: y.shift(1).ewm(span=5).mean(),
        }

        g = {"const": lambda x: x + 10, "trend": lambda x: -x}

        weights = np.power(np.arange(start=0, stop=1, step=1 / len(endog)), 2)

        return data, f, g, weights

    def test_model_fit(self):
        data, f, g, weights = self.generate_input()

        y_train = data[self.lbl_value]
        x_train = data.iloc[:, 1:]

        model = HandCraftedLinearModel(endog_transform=f, exog_transform=g)
        model.fit(y=y_train, X=x_train, weights=weights)

        parameters = model.get_parameters()

        keys = set(f.keys())
        for key in g.keys():
            keys.update({"{} {}".format(key, col) for col in x_train.columns})

        # Some random test. No good logic here
        assert set(parameters.index) == keys
        assert parameters.isna().sum() == 0
        assert set(model.summary()[model.lbl_params].keys()) == keys

    def test_model_prediction(self):
        data, f, g, weights = self.generate_input()

        num_steps = 5
        model = HandCraftedLinearModel(endog_transform=f, exog_transform=g)
        y_train = data.loc[data.index[:-num_steps], self.lbl_value]
        x_train = data.iloc[:-num_steps, 1:]
        x_test = data.iloc[-num_steps:, 1:]
        weights_train = weights[:-num_steps]

        model.fit(y=y_train, X=x_train, weights=weights_train)
        forecast = model.predict(num_steps=num_steps, X=x_test)

        assert isinstance(forecast, pd.DataFrame)
        assert forecast.shape[0] == num_steps
        assert forecast.columns[0] == self.lbl_value
        assert forecast.isna().sum().sum() == 0
        assert forecast.index.name == self.lbl_date
        assert isinstance(forecast.index, pd.DatetimeIndex)



from __future__ import annotations

from typing import List

import pandas as pd
from statsmodels.regression.linear_model import OLS
from statsmodels.tsa.statespace.sarimax import SARIMAX
from statsmodels.tsa.tsatools import add_trend

from hcl_model.model_base import ModelBase


class SARIMAXModel(ModelBase):
    """
    SARIMAX model

    Wrapper around `statsmodels implementation
    <http://www.statsmodels.org/stable/generated/statsmodels.tsa.statespace.sarimax.SARIMAX.html#statsmodels.tsa.statespace.sarimax.SARIMAX>`_.

    See documentation of `statsmodels.tsa.statespace.sarimax.SARIMAX`
    """

    def __init__(
        self,
        order: tuple = (0, 0, 0),
        seasonal_order: tuple = (0, 0, 0, 0),
        trend: str = "c",
        enforce_stationarity: bool = True,
    ) -> None:
        super().__init__()
        self._order = order
        self._seasonal_order = seasonal_order
        self._trend = trend
        self._enforce_stationarity = enforce_stationarity
        self._trend_fit = None

    def _fit(self, **kwargs) -> None:
        self._y_train = self._remove_trend(self._y_train)
        self._fit_results = SARIMAX(
            self._y_train,
            exog=self._x_train,
            order=self._order,
            seasonal_order=self._seasonal_order,
            enforce_stationarity=self._enforce_stationarity,
        ).fit(disp=False)

    def _predict(
        self, num_steps: int, X: pd.DataFrame = None, quantile_levels: List[float] = None, num_simulations: int = None
    ) -> pd.DataFrame:
        return (
            self._fit_results.get_forecast(steps=num_steps, exog=X)
            .predicted_mean.rename(self._get_endog_name())
            .to_frame()
            .rename_axis(index=self._y_train.index.name)
        )

    def _simulate(self, num_steps: int, num_simulations: int, X: pd.DataFrame = None, **kwargs) -> pd.DataFrame:
        self._y_train = self._remove_trend(self._y_train)
        sim_model = SARIMAX(
            pd.Series(index=X.index, dtype=float),
            exog=X,
            order=self._order,
            seasonal_order=self._seasonal_order,
            enforce_stationarity=self._enforce_stationarity,
        )
        sim_model = sim_model.filter(params=self._fit_results.params)
        # TODO: check simulation output for different model. I am not sure it is correct without initial_sate.
        return pd.DataFrame({i: sim_model.simulate(num_steps) for i in range(num_simulations)})

    def _get_aic(self) -> float:
        return self._fit_results.aic

    def _get_fitted_values(self) -> pd.Series:
        return self._fit_results.fittedvalues

    def _get_residuals(self) -> pd.Series:
        return self._fit_results.resid

    def _compute_prediction_quantiles(
        self, num_steps: int, quantile_levels: List[float] = None, X: pd.DataFrame = None, **kwargs
    ) -> pd.DataFrame:
        forecast = self._fit_results.get_forecast(steps=num_steps, exog=X)
        out = dict()
        for alpha, q_name in zip(quantile_levels, self.get_quantile_names(quantile_levels)):
            if alpha < 50:
                out[q_name] = forecast.conf_int(alpha=2 * alpha / 100).iloc[:, 0]
            else:
                out[q_name] = forecast.conf_int(alpha=2 * (100 - alpha) / 100).iloc[:, 1]
        return pd.DataFrame(out).rename_axis(index=self._y_train.index.name)

    def _remove_trend(self, endog: pd.Series) -> pd.Series:
        if self._trend == "n":
            return endog
        else:
            name = endog.name
            trend = add_trend(self._y_train, trend=self._trend, prepend=False)
            self._trend_fit = OLS(endog, trend.iloc[:, 1:]).fit()
            endog -= self._trend_fit.fittedvalues
            return endog.rename(name)

    def _add_trend(self, df: pd.DataFrame) -> pd.DataFrame:
        if self._trend == "n":
            return df
        else:
            exog = add_trend(
                pd.concat([self._y_train, df[self._y_train.name]]),
                trend=self._trend,
                prepend=False,
                has_constant="add",
            )
            return df.apply(lambda x: x + self._trend_fit.predict(exog.iloc[-df.shape[0] :, 1:]))



from __future__ import annotations

from typing import Union

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin

X_TYPE = Union[pd.Series, pd.DataFrame]


class TruncateTransformer(BaseEstimator, TransformerMixin):
    """Truncate data.

    Leave only a specific number of past weeks.
    """

    def __init__(self, full_weight_weeks: int, truncate: bool = True) -> None:
        self.truncate = truncate
        self.full_weight_weeks = full_weight_weeks

    def fit(self, X: X_TYPE, y: pd.Series = None) -> TruncateTransformer:
        return self

    def transform(self, X: X_TYPE) -> X_TYPE:
        if self.truncate and self.full_weight_weeks > 0:
            return X[X.index > X.index.max() - pd.DateOffset(weeks=self.full_weight_weeks, days=1)]
        else:
            return X



from __future__ import annotations

import pandas as pd
import pytest
from pandas._testing import assert_series_equal
from sklearn.base import BaseEstimator
from sklearn.exceptions import NotFittedError

from hcl_model.transformers.estimator_to_transformer import EstimatorToTransformer


class ModelForTests(BaseEstimator):
    def __init__(self) -> None:
        super().__init__()
        self.indicator = "A"

    def fit(self, X: pd.DataFrame, y: pd.Series = None) -> ModelForTests:
        self.indicator = "B"
        return self

    def predict(self, X: pd.DataFrame) -> pd.Series:
        return pd.Series([1.0, 5.0])


def test_estimator_to_transformer() -> None:
    custom_model = ModelForTests()

    assert custom_model.indicator == "A"

    sklearn_model = EstimatorToTransformer(model=custom_model)
    x_data = pd.DataFrame({"x": [1.0, 3.0]})

    with pytest.raises(NotFittedError, match="instance is not fitted yet"):
        sklearn_model.transform(X=x_data)

    sklearn_model.fit(X=x_data, y=pd.Series([4.0]))

    assert custom_model.indicator == "B"

    assert_series_equal(sklearn_model.transform(X=x_data), pd.Series([1.0, 5.0]))



import numpy as np
import pandas as pd

from hcl_model.utils.decayed_weights import decayed_weights


def test_decayed_weights() -> None:
    for nobs in [5, 10, 99, 100, 101]:
        endog = pd.Series(
            np.random.normal(size=nobs),
            index=pd.date_range("2018-01-01", periods=nobs, freq="W-FRI"),
        )

        for weights in [
            decayed_weights(endog=endog),
            decayed_weights(endog=endog, full_weight_obs=10),
            decayed_weights(endog=endog, full_weight_obs=nobs + 10),
            decayed_weights(endog=endog, full_weight_obs=10, downweight_order=0),
            decayed_weights(endog=endog, full_weight_obs=10, downweight_order=1),
            decayed_weights(endog=endog, full_weight_obs=10, downweight_order=2),
            decayed_weights(endog=endog, full_weight_obs=10, downweight_order=3),
        ]:
            assert isinstance(weights, np.ndarray)
            assert weights.shape[0] == endog.shape[0]
            assert (weights >= 0).all()
            assert (weights <= 1).all()

        weights = decayed_weights(endog=endog, full_weight_obs=nobs + 10)

        assert (weights == 1).all()

        weights = decayed_weights(endog=endog, full_weight_obs=10, downweight_order=0)

        assert (weights == 1).all()



import pandas as pd
from workalendar.core import Calendar
from workalendar.registry import registry

from hcl_model.labels import LabelsCommon, LabelsHolidaysOut


class CalendarReader:
    lbl = LabelsCommon()
    lbl_holidays_out = LabelsHolidaysOut()

    def get_holidays(self, holiday_name: str, country_code: str, from_year: int, to_year: int) -> pd.DataFrame:
        """Import holidays.

        The reader uses 'workalendar' library available here: https://pypi.org/project/workalendar/.

        :param holiday_name: holiday name
        :param country_code: two letter country code, e.g. `DE` or `CN`
        :param from_year: the year from which to start the calendar
        :param to_year: the year to end the calendar
        :return: DataFrame with holiday dates for specific countries.

        Typical output:

        .. code-block:: python

                                 holiday country
            date
            2010-04-05     Easter Monday      DE
            2002-04-01     Easter Monday      DE
            2030-04-22     Easter Monday      DE
            2009-04-13     Easter Monday      DE

        The DataFrame is indexed by date of the holiday

        Two columns:

        - holiday: name of the holiday
        - country: iso code of the country

        """
        cal = self._get_calendar(country_code)

        # Go over all holidays in each year find those that match the request
        holiday_list = [x[0] for year in range(from_year, to_year) for x in cal.holidays(year) if holiday_name == x[1]]
        if not holiday_list:
            raise RuntimeError('Holiday "{}" was not found in "{}" calendar!'.format(holiday_name, country_code))

        # Create a DataFrame with date as index, country and holiday name as columns
        return pd.DataFrame(
            {
                self.lbl_holidays_out.lbl_holiday: holiday_name,
                self.lbl_holidays_out.lbl_country: country_code,
            },
            index=pd.Index(pd.to_datetime(holiday_list), name=self.lbl.date),
        )

    @staticmethod
    def _get_calendar(country_code: str = "CN") -> Calendar:
        """Get calendar object given two-letter code of a country.

        :param country_code: two letter country code, e.g. `DE` or `CN`
        :return: Calendar class for specific country
        """
        if country_code in registry.region_registry.keys():
            cal_class = registry.get(country_code)
            return cal_class()
        else:
            raise RuntimeError("{} is an unknown country code!".format(country_code))



from __future__ import annotations

from typing import Union

import numpy as np
import pandas as pd
import ruptures as rpt
from sklearn.base import BaseEstimator, TransformerMixin

X_TYPE = Union[pd.Series, np.ndarray]


class TargetStructuralBreakCorrectionTransformer(BaseEstimator, TransformerMixin):
    def __init__(self, structural_break_correction: bool = True) -> None:
        self.structural_break_correction = structural_break_correction

    def fit(self, X: X_TYPE, y: pd.Series = None) -> TargetStructuralBreakCorrectionTransformer:
        return self

    def transform(self, X: X_TYPE) -> X_TYPE:
        if self.structural_break_correction:
            if isinstance(X, np.ndarray):
                return self._get_series_without_structural_breaks(signal=pd.Series(X.flatten())).values
            else:
                return self._get_series_without_structural_breaks(signal=X)
        else:
            return X

    def _get_series_without_structural_breaks(self, signal: pd.Series) -> pd.Series:
        change_points = self._get_change_points(y=signal)
        if len(change_points) <= 1:
            return signal
        else:
            change_points = np.concatenate(([0], change_points))
            current_signal = signal[change_points[-2] : change_points[-1]]
            level_current = current_signal.median()
            variability_current = current_signal.std()
            for past, current in zip(change_points[:-1], change_points[1:]):
                variability_past = signal[past:current].std()
                signal[past:current] = self._adjust_level(y=signal[past:current], level_current=level_current)
                signal[past:current] = self._adjust_variability(
                    y=signal[past:current], variability_current=variability_current, variability_past=variability_past
                )
            return signal

    @staticmethod
    def _adjust_variability(y: pd.Series, variability_current: float, variability_past: float) -> pd.Series:
        if variability_past == 0:
            return y
        else:
            adjust_factor = variability_current / variability_past
            return (y - y.mean()) * adjust_factor + y.mean()

    @staticmethod
    def _adjust_level(y: pd.Series, level_current: float) -> pd.Series:
        return y + level_current - y.median()

    @staticmethod
    def _get_change_points(y: pd.Series) -> np.array:
        return np.array(rpt.KernelCPD(kernel="rbf", jump=1, min_size=26).fit(y.values).predict(pen=10))



import pandas as pd

from hcl_model.model_sarimax import SARIMAXModel
from tests.test_model_common import TestModelCommon


class TestSARIMAX(TestModelCommon):
    def test_model_fit(self):

        endog, exog = self.generate_data()

        model = SARIMAXModel(trend="n")
        y_train = endog[self.lbl_value]
        model.fit(y=y_train, X=None)
        parameters = model.get_parameters()

        assert list(parameters.index) == ["sigma2"]
        assert model._trend_fit is None
        assert set(model.summary()[model.lbl_params].keys()) == {"sigma2"}

        model = SARIMAXModel(trend="c")
        model.fit(y=y_train, X=None)
        parameters = model.get_parameters()

        assert list(parameters.index) == ["sigma2"]
        assert set(model._trend_fit.params.index.values) == {"const"}
        # trend is extracted before fitting SARIMAX, hence no 'const' among parameters
        assert set(model.summary()[model.lbl_params].keys()) == {"sigma2"}

        model = SARIMAXModel(trend="ct")
        model.fit(y=y_train, X=None)
        parameters = model.get_parameters()

        assert list(parameters.index) == ["sigma2"]
        assert set(model._trend_fit.params.index.values) == {"const", "trend"}
        assert set(model.summary()[model.lbl_params].keys()) == {"sigma2"}

        model = SARIMAXModel(trend="t")
        model.fit(y=y_train, X=None)
        parameters = model.get_parameters()

        assert list(parameters.index) == ["sigma2"]
        assert set(model._trend_fit.params.index.values) == {"trend"}
        assert set(model.summary()[model.lbl_params].keys()) == {"sigma2"}

        model = SARIMAXModel(trend="n")
        model.fit(y=y_train, X=exog)
        parameters = model.get_parameters()

        assert list(parameters.index) == ["const", "time", "sigma2"]
        assert model._trend_fit is None
        assert set(model.summary()[model.lbl_params].keys()) == {"const", "time", "sigma2"}

    def test_model_prediction(self):
        endog, exog = self.generate_data()
        model = SARIMAXModel(trend="ct")
        num_steps = 10
        y_train = endog.loc[endog.index[:-num_steps], self.lbl_value]
        x_train = exog.loc[endog.index[:-num_steps]]
        x_test = exog.loc[endog.index[-num_steps:]]
        model.fit(y=y_train, X=x_train)
        forecast = model.predict(num_steps=num_steps, X=x_test)

        assert isinstance(forecast, pd.DataFrame)
        assert forecast.shape[0] == num_steps
        assert forecast.columns[0] == self.lbl_value
        assert forecast.index.name == self.lbl_date
        assert isinstance(forecast.index, pd.DatetimeIndex)

    def test_model_simulation(self):
        endog, exog = self.generate_data()
        model = SARIMAXModel(trend="ct")
        num_steps = 10
        num_simulations = 5

        y_train = endog.loc[endog.index[:-num_steps], self.lbl_value]
        x_train = exog.loc[endog.index[:-num_steps]]
        x_test = exog.loc[endog.index[-num_steps:]]
        model.fit(y=y_train, X=x_train)
        simulations = model.simulate(num_steps=num_steps, num_simulations=num_simulations, X=x_test)

        assert isinstance(simulations, pd.DataFrame)
        assert simulations.shape == (num_steps, num_simulations)
        assert simulations.index.name == self.lbl_date
        assert isinstance(simulations.index, pd.DatetimeIndex)

    def test_model_percentiles(self):
        endog, exog = self.generate_data()
        model = SARIMAXModel(trend="ct")
        num_steps = 10
        num_simulations = 5
        quantile_levels = [5, 95]

        y_train = endog.loc[endog.index[:-num_steps], self.lbl_value]
        x_train = exog.loc[endog.index[:-num_steps]]
        model.fit(y=y_train, X=x_train)
        x_test = exog.loc[endog.index[-num_steps:]]
        forecast = model.predict(
            num_steps=num_steps, quantile_levels=quantile_levels, num_simulations=num_simulations, X=x_test
        )

        assert isinstance(forecast, pd.DataFrame)
        assert forecast.shape == (num_steps, len(quantile_levels) + 1)
        assert forecast.columns[0] == self.lbl_value
        assert forecast.index.name == self.lbl_date
        assert isinstance(forecast.index, pd.DatetimeIndex)

    def test_model_summary(self):
        endog, exog = self.generate_data()

        model = SARIMAXModel(trend="n")
        model.fit(y=endog[self.lbl_value], X=None)

        assert set(model.summary().index) >= {
            model.lbl_aic,
            model.lbl_r2,
            model.lbl_mape,
            model.lbl_resid_mean,
            model.lbl_resid_std,
            model.lbl_resid_skewness,
            model.lbl_resid_kurtosis,
            model.lbl_params,
        }



import numpy as np
import pandas as pd

from hcl_model.transformers.calendar import CalendarTransformer
from hcl_model.utils.construct_calendar_exogenous import construct_calendar_exogenous


class TestConstructCalendarExogenous:

    # countries
    lbl_country_code_tw = "TW"
    lbl_country_code_de = "DE"

    # holidays
    lbl_cny = "Chinese New Year"
    lbl_easter_monday = "Easter Monday"
    lbl_new_year = "New year"
    lbl_cgw_tw = "National Day/Double Tenth Day"

    lbl_holiday_name = "holiday_name"
    lbl_country_code = "country_code"
    lbl_lags = "lags"
    lbl_add_td = "add_td"

    def test_construct_calendar_exogenous(self):
        np.random.seed(42)
        nobs, num_steps = 100, 26
        endog = pd.Series(
            np.random.random(nobs) + 100,
            index=pd.date_range("2017-02-01", periods=nobs, freq="W-FRI"),
        )
        lbl_endog = "endog"
        df = pd.DataFrame({lbl_endog: endog})
        lbl_dummy = CalendarTransformer.lbl_auto_dummy

        weeks = [1, 5, 10, 50]
        for week in weeks:
            df.loc[df.index.map(lambda x: x.isocalendar()[1]) == week, lbl_endog] += 100
            df[lbl_dummy.format(week)] = 0.0
            df[lbl_dummy.format(week + 1)] = 0.0
            df.loc[
                df.index.map(lambda x: x.isocalendar()[1]) == week,
                lbl_dummy.format(week),
            ] = 1
            df.loc[
                df.index.map(lambda x: x.isocalendar()[1]) == week + 1,
                lbl_dummy.format(week + 1),
            ] = 1

        endog = df[lbl_endog]

        exog = construct_calendar_exogenous(endog=endog, num_steps=num_steps)
        assert exog.shape == (nobs + num_steps, 1)

        for trend in ["c", "t", "ct", "ctt"]:
            exog = construct_calendar_exogenous(endog=endog, num_steps=num_steps, trend=trend)
            assert exog.shape == (nobs + num_steps, len(trend))

        for num in [1, 2, 3]:
            exog = construct_calendar_exogenous(endog=endog, num_steps=num_steps, splines_df=num)
            assert exog.shape == (nobs + num_steps, num + 1)

        holidays = [
            {
                self.lbl_holiday_name: self.lbl_cny,
                self.lbl_country_code: self.lbl_country_code_tw,
                self.lbl_lags: [1],
            },
            {
                self.lbl_holiday_name: self.lbl_new_year,
                self.lbl_country_code: self.lbl_country_code_de,
                self.lbl_lags: [-1, 0, 2],
            },
            {
                self.lbl_holiday_name: self.lbl_easter_monday,
                self.lbl_country_code: self.lbl_country_code_de,
                self.lbl_lags: [-1, -2],
                self.lbl_add_td: "59 D",
            },
            {
                self.lbl_holiday_name: self.lbl_cgw_tw,
                self.lbl_country_code: self.lbl_country_code_tw,
            },
        ]

        for holiday in holidays:
            exog = construct_calendar_exogenous(endog=endog, num_steps=num_steps, holidays=[holiday])
            if self.lbl_lags in holiday.keys():
                ncols = len(holiday[self.lbl_lags])
            else:
                ncols = 5
            assert exog.shape == (nobs + num_steps, ncols + 1)

        exog = construct_calendar_exogenous(
            endog=endog,
            num_steps=num_steps,
            auto_dummy_max_number=len(weeks) * 2,
            auto_dummy_threshold=3,
        )

        assert exog.shape == (nobs + num_steps, len(weeks) + 1)

        holiday = {
            self.lbl_holiday_name: self.lbl_new_year,
            self.lbl_country_code: self.lbl_country_code_de,
            self.lbl_lags: [-1, 0, 1],
        }

        exog = construct_calendar_exogenous(
            endog=endog,
            num_steps=num_steps,
            holidays=[holiday],
            auto_dummy_max_number=len(weeks) * 2,
            auto_dummy_threshold=3,
        )

        # test that one column is dropped since automatic dummy coincides with NY
        assert len(holiday[self.lbl_lags]) + len(weeks) == len(exog.columns)



import pandas as pd
import pytest

from hcl_model.utils.check_x_y import check_X_y


def test_check_X_y() -> None:
    index = pd.date_range(start="2021-01-01", periods=3, freq="D", name="date")
    X_raw = pd.DataFrame({"a": [1, 2, 4]}, index=index)
    y_raw = pd.Series([3, 5, 7], index=index, name="some name")

    X, y = check_X_y(X=X_raw, y=y_raw)
    pd.testing.assert_frame_equal(X, X_raw)
    pd.testing.assert_series_equal(y, y_raw)

    X, y = check_X_y(X=X_raw, y=y_raw.values)
    pd.testing.assert_frame_equal(X, X_raw)
    pd.testing.assert_series_equal(y, y_raw.rename("value"))

    X, y = check_X_y(X=None, y=y_raw)
    assert X is None
    pd.testing.assert_series_equal(y, y_raw)

    error_msg = "Passing ndarray for y and None for X is not supported"
    with pytest.raises(TypeError, match=error_msg):
        check_X_y(X=None, y=y_raw.values)

    error_msg = r"Length of values \(\d\) does not match length of index \(\d\)"
    with pytest.raises(ValueError, match=error_msg):
        check_X_y(X=X_raw, y=y_raw.values[1:])



from workalendar.core import Calendar

from hcl_model.utils.calendar_reader import CalendarReader


class TestCalendarReader:

    from_year = 2005
    to_year = 2030
    lbl_new_year = "New year"

    lbl_country_code_de = "DE"
    lbl_country_code_cn = "CN"
    lbl_country_code_tw = "TW"

    def test_get_holidays(self):
        cal_reader = CalendarReader()

        df = cal_reader.get_holidays(self.lbl_new_year, self.lbl_country_code_de, self.from_year, self.to_year)

        assert df.shape[0] == self.to_year - self.from_year
        assert set(df.values.ravel()) == {self.lbl_new_year, self.lbl_country_code_de}
        assert df.index.year[0] == self.from_year
        assert df.index.year[-1] == self.to_year - 1

    def test_get_calendar(self):
        cal_reader = CalendarReader()

        assert isinstance(cal_reader._get_calendar(self.lbl_country_code_de), Calendar)



from __future__ import annotations

from typing import Optional

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted


class EstimatorToTransformer(BaseEstimator, TransformerMixin):
    """Convert sklearn estimator to transformer.

    Effectively, it replaces `predict` method with `transform`,
    so that the object is usable as an intermediate step in `Pipeline`.

    :param model: object with fit and predict methods
    """

    def __init__(self, model: BaseEstimator) -> None:
        self.model = model

    def fit(self, X: pd.DataFrame, y: pd.Series = None) -> EstimatorToTransformer:
        self.model.fit(X=X, y=y)
        self.model_ = self.model
        return self

    def transform(self, X: pd.DataFrame) -> Optional[pd.Series]:
        check_is_fitted(self)
        return self.model_.predict(X=X)



from typing import Optional, Tuple, Union

import numpy as np
import pandas as pd


def check_X_y(X: Optional[pd.DataFrame], y: Union[pd.Series, np.ndarray]) -> Tuple[Optional[pd.DataFrame], pd.Series]:
    X = X.copy() if X is not None else None
    if isinstance(y, np.ndarray) & (X is not None):
        y = pd.Series(y, index=X.index, name="value")
    elif isinstance(y, np.ndarray):
        raise TypeError(
            "Passing ndarray for y and None for X is not supported. "
            "Please, consider converting y to Series with DatetimeIndex."
        )
    else:
        y = y.copy()
    return X, y



import numpy as np
import pandas as pd


def decayed_weights(endog: pd.Series, full_weight_obs: int = 52, downweight_order: int = 2) -> np.ndarray:
    """Construct weights starting to decay at `start_from` with polynomial order.

    :param endog: pd.Series with index of datetime type in consecutive order.
    :param full_weight_obs: number of observations from the end that have full weight (1).
    :param downweight_order: polynomial order of decrease in weights.
    :return: numpy 1d array of floats between 0 and 1 with same length as endog.
    """
    len_weights = endog.size

    if len_weights <= full_weight_obs:
        return np.ones(len_weights)
    else:
        downweighted = np.power(
            np.linspace(start=0, stop=1, num=len_weights - full_weight_obs),
            downweight_order,
        )
        return np.concatenate((downweighted, np.ones(full_weight_obs)), axis=None)



from typing import List

import pandas as pd


def get_duplicate_columns(df: pd.DataFrame) -> List[str]:
    """Get a list of duplicate columns.

    https://thispointer.com/how-to-find-drop-duplicate-columns-in-a-dataframe-python-pandas/

    It will iterate over all the columns in dataframe and find the columns whose contents are duplicate.

    :param df: Dataframe object
    :return: List of columns whose contents are duplicates.
    """
    duplicate_column_names = set()
    # Iterate over all the columns in dataframe
    for x in range(df.shape[1]):
        # Select column at xth index.
        col = df.iloc[:, x]
        # Iterate over all the columns in DataFrame from (x+1)th index till end
        for y in range(x + 1, df.shape[1]):
            # Select column at yth index.
            other_col = df.iloc[:, y]
            # Check if two columns at x 7 y index are equal
            if col.equals(other_col):
                duplicate_column_names.add(df.columns.values[y])

    return list(duplicate_column_names)



import numpy as np
import pandas as pd
import pytest

from hcl_model.transformers.calendar import CalendarTransformer


class TestCalendarTransformer:
    # countries
    lbl_country_code_tw = "TW"
    lbl_country_code_de = "DE"

    # holidays
    lbl_cny = "Chinese New Year"
    lbl_easter_monday = "Easter Monday"
    lbl_new_year = "New year"
    lbl_cgw_tw = "National Day/Double Tenth Day"

    @staticmethod
    def get_xreg_example() -> pd.DataFrame:
        return pd.DataFrame(
            {"x1": np.random.normal(0, 1, 53 * 3)},
            index=pd.date_range(start="2015-01-01", freq="7 D", periods=53 * 3),
        )

    def test_add_periodic_splines(self):
        xreg = self.get_xreg_example()
        cal_transformer = CalendarTransformer()

        degrees_of_freedom = 3
        df_default = cal_transformer.add_periodic_splines(df=xreg, degrees_of_freedom=degrees_of_freedom)

        assert df_default.shape == (xreg.shape[0], degrees_of_freedom + xreg.shape[1])
        assert np.sum(df_default.isnull()).sum() == 0

        df = cal_transformer.add_periodic_splines(df=xreg, degrees_of_freedom=degrees_of_freedom, input_level="isoweek")

        assert df.shape == (xreg.shape[0], degrees_of_freedom + xreg.shape[1])
        assert np.sum(df.isnull()).sum() == 0
        pd.testing.assert_frame_equal(df, df_default)

        df = cal_transformer.add_periodic_splines(df=xreg, degrees_of_freedom=degrees_of_freedom, input_level="month")

        assert df.shape == (xreg.shape[0], degrees_of_freedom + xreg.shape[1])
        assert np.sum(df.isnull()).sum() == 0

        with pytest.raises(ValueError):
            cal_transformer.add_periodic_splines(df=xreg, degrees_of_freedom=degrees_of_freedom, input_level="day")

        with pytest.raises(ValueError) as error:
            wrong_input_level = "wrong_input_level"
            cal_transformer.add_periodic_splines(
                df=xreg,
                degrees_of_freedom=degrees_of_freedom,
                input_level=wrong_input_level,
            )

        assert wrong_input_level in str(error)

    def test_add_holiday_triangles(self):
        xreg = self.get_xreg_example()
        cal_transformer = CalendarTransformer()

        lh_factors = [1.0, 2.0]
        rh_factors = [7.0]

        tasks = [
            (self.lbl_cny, self.lbl_country_code_tw),
            (self.lbl_new_year, self.lbl_country_code_de),
            (self.lbl_easter_monday, self.lbl_country_code_de),
            (self.lbl_cgw_tw, self.lbl_country_code_tw),
        ]

        for holiday, country in tasks:
            df1 = cal_transformer.add_holiday_triangles(
                xreg, holiday, country, lh_factors=lh_factors, rh_factors=rh_factors
            )
            df2 = cal_transformer.add_holiday_triangles(xreg, holiday, country, merge_factors=True)

            assert df1.shape == (xreg.shape[0], xreg.shape[1] + 2)
            assert df2.shape == (xreg.shape[0], xreg.shape[1] + 1)

            assert set(df1.iloc[:, -2:].values.ravel()) <= set([0.0] + lh_factors + rh_factors)

    def test_add_holiday_dummies(self):
        xreg = self.get_xreg_example()
        cal_transformer = CalendarTransformer()

        xreg_cny = (
            cal_transformer.add_holiday_dummies(xreg, self.lbl_cny, self.lbl_country_code_tw, lags=[1]),
            [1],
        )
        xreg_ny = (
            cal_transformer.add_holiday_dummies(xreg, self.lbl_new_year, self.lbl_country_code_de, lags=[-1, 0, 2]),
            [-1, 0, 2],
        )
        xreg_cc = (
            cal_transformer.add_holiday_dummies(
                xreg,
                self.lbl_easter_monday,
                self.lbl_country_code_de,
                add_td="59 D",
                lags=[-1, -2],
            ),
            [-1, -2],
        )
        xreg_cgw = (
            cal_transformer.add_holiday_dummies(xreg, self.lbl_cgw_tw, self.lbl_country_code_tw),
            None,
        )

        for df, lags in [xreg_cny, xreg_cc, xreg_ny, xreg_cgw]:

            if lags is None:
                lags_len = 5  # default is lags=(-1,-2,0,1,2)
            else:
                lags_len = len(lags)

            assert np.sum(df.isnull()).sum() == 0
            assert df.shape[1] == lags_len + 1
            assert df.shape[0] == xreg.shape[0]
            assert set(np.ravel(df.iloc[:, 1:].values)) == {0.0, 1.0}


class TestAddAutomaticSeasonalDummies:
    def test_add_automatic_seasonal_dummies(self):
        np.random.seed(42)
        nobs = 200
        lbl_endog = "endog"
        lbl_date = "date"
        lbl_dummy = CalendarTransformer.lbl_auto_dummy
        df = pd.DataFrame(
            {lbl_endog: np.random.random(nobs) + 100},
            index=pd.DatetimeIndex(pd.date_range("2017-01-01", periods=nobs, freq="W-FRI"), name=lbl_date),
        )
        weeks = {2: 10, 5: 5, 10: 20, 50: 7}
        for week, change in weeks.items():
            df.loc[df.index.map(lambda x: x.isocalendar()[1]) == week, lbl_endog] += change
            df[lbl_dummy.format(week)] = 0.0
            df.loc[
                df.index.map(lambda x: x.isocalendar()[1]) == week,
                lbl_dummy.format(week),
            ] = 1

        df_result = CalendarTransformer.add_automatic_seasonal_dummies(
            df=df[[lbl_endog]],
            var_name=lbl_endog,
            threshold=3,
            lim_num_dummies=len(weeks) * 2,
        )

        pd.testing.assert_frame_equal(df.sort_index(axis=1), df_result.sort_index(axis=1))

        df_result = CalendarTransformer.add_automatic_seasonal_dummies(
            df=df[[lbl_endog]], var_name=lbl_endog, threshold=3, lim_num_dummies=2
        )

        correct_cols = [lbl_endog, lbl_dummy.format(10), lbl_dummy.format(2)]

        pd.testing.assert_frame_equal(
            df.filter(items=correct_cols).sort_index(axis=1),
            df_result.sort_index(axis=1),
        )

        df_result = CalendarTransformer.add_automatic_seasonal_dummies(
            df=df[[lbl_endog]], var_name=lbl_endog, threshold=3, lim_num_dummies=4
        )
        correct_cols = [
            lbl_endog,
            lbl_dummy.format(10),
            lbl_dummy.format(50),
            lbl_dummy.format(2),
            lbl_dummy.format(5),
        ]

        pd.testing.assert_frame_equal(
            df.filter(items=correct_cols).sort_index(axis=1),
            df_result.sort_index(axis=1),
        )

        df_result = CalendarTransformer.add_automatic_seasonal_dummies(
            df=df[[lbl_endog]], var_name=lbl_endog, threshold=3, lim_num_dummies=6
        )
        correct_cols = [
            lbl_endog,
            lbl_dummy.format(10),
            lbl_dummy.format(2),
            lbl_dummy.format(5),
            lbl_dummy.format(50),
        ]

        pd.testing.assert_frame_equal(
            df.filter(items=correct_cols).sort_index(axis=1),
            df_result.sort_index(axis=1),
        )

        num_dummies = 0
        for quantile in np.linspace(0.99, 0.01, num=10):
            df_result = CalendarTransformer.add_automatic_seasonal_dummies(
                df=df[[lbl_endog]],
                var_name=lbl_endog,
                threshold=quantile,
                lim_num_dummies=len(weeks) * 2,
            )
            num_dummies_new = len(df_result.columns) - 1
            assert num_dummies_new >= num_dummies
            num_dummies = num_dummies_new

        assert num_dummies == len(weeks) * 2

    def test_add_automatic_seasonal_dummies_raises(self):
        nobs = 200
        lbl_endog = "endog"
        lbl_date = "date"
        freq = "M"
        df = pd.DataFrame(
            {lbl_endog: np.random.random(nobs)},
            index=pd.DatetimeIndex(pd.date_range("2017-01-01", periods=nobs, freq=freq), name=lbl_date),
        )

        with pytest.raises(RuntimeError) as e:
            CalendarTransformer.add_automatic_seasonal_dummies(
                df=df[[lbl_endog]], var_name=lbl_endog, threshold=0.5, lim_num_dummies=8
            )

            assert freq in str(e)

    def test_add_automatic_seasonal_dummies_empty(self):
        nobs = 200
        lbl_endog = "endog"
        lbl_date = "date"
        df = pd.DataFrame(
            {lbl_endog: np.ones(nobs)},
            index=pd.DatetimeIndex(pd.date_range("2017-01-01", periods=nobs, freq="W-FRI"), name=lbl_date),
        )

        df_result = CalendarTransformer.add_automatic_seasonal_dummies(
            df=df[[lbl_endog]], var_name=lbl_endog, threshold=3, lim_num_dummies=8
        )

        pd.testing.assert_frame_equal(df[[lbl_endog]], df_result)



import numpy as np
import pandas as pd
import pytest

from hcl_model.utils.smooth import smooth_series


def test_smooth_series() -> None:
    y_original = pd.Series(range(50))
    y_smoothed = smooth_series(y=y_original, window=10, quantile=0.5, ewm_alpha=0.1)
    assert len(y_original) == len(y_smoothed)

    y_original = pd.Series(np.zeros(100))
    y_smoothed = smooth_series(y=y_original, window=10, quantile=0.5, ewm_alpha=0.1)
    pd.testing.assert_series_equal(y_original, y_smoothed)

    y_original = pd.Series([0, 1] * 50)
    y_smoothed = smooth_series(y=y_original, window=10, quantile=0.25, ewm_alpha=0.1)
    assert sum(y_smoothed == 0) == len(y_smoothed)
    assert sum(y_smoothed == 1) == 0

    y_smoothed = smooth_series(y=y_original, window=10, quantile=0.5, ewm_alpha=0.1)
    assert sum(y_smoothed < y_original) == sum(y_original == 1)
    assert sum(y_smoothed > y_original) == sum(y_original == 0)

    y_smoothed = smooth_series(y=y_original, window=10, quantile=0.75, ewm_alpha=0.1)
    assert sum(y_smoothed == 1) == len(y_smoothed)
    assert sum(y_smoothed == 0) == 0

    with pytest.raises(ValueError):
        smooth_series(y=y_original, window=-10, quantile=0.5, ewm_alpha=0.1)

    with pytest.raises(ValueError):
        smooth_series(y=y_original, window=10, quantile=-0.5, ewm_alpha=0.1)

    with pytest.raises(ValueError):
        smooth_series(y=y_original, window=10, quantile=0.5, ewm_alpha=-0.1)



from __future__ import annotations

from typing import Union

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin

from hcl_model.utils.smooth import smooth_series
from hcl_model.utils.string_enum import StringEnum

X_TYPE = Union[pd.Series, np.ndarray]


class CorrectOutliersMethodNames(StringEnum):
    nothing = "nothing"
    interpolate = "interpolate"
    trim = "trim"


class TargetOutlierCorrectionTransformer(BaseEstimator, TransformerMixin):
    """Target outlier correction."""

    def __init__(
        self,
        outlier_correction_method: CorrectOutliersMethodNames,
        six_sigma_multiplier: float,
        smoothing_window: int = 52,
        ewm_alpha: float = 0.1,
    ) -> None:
        self.six_sigma_multiplier = six_sigma_multiplier
        self.smoothing_window = smoothing_window
        self.ewm_alpha = ewm_alpha
        self.outlier_correction_method = outlier_correction_method

    def fit(self, X: X_TYPE, y: pd.Series = None) -> TargetOutlierCorrectionTransformer:
        return self

    def transform(self, X: X_TYPE) -> X_TYPE:
        if self.outlier_correction_method == CorrectOutliersMethodNames.nothing:
            return X
        else:
            if isinstance(X, np.ndarray):
                return self._correct_outliers(series=pd.Series(X.flatten())).values
            else:
                return self._correct_outliers(series=X)

    def _correct_outliers(self, series: pd.Series) -> pd.Series:
        is_too_high = self.get_idx_too_high(series=series)
        is_too_low = self.get_idx_too_low(series=series)
        series_corrected = series.copy()
        if sum(is_too_high) + sum(is_too_low) > 0:
            if self.outlier_correction_method == CorrectOutliersMethodNames.trim:
                series_corrected[is_too_low] = self.get_lower_limit(series=series)[is_too_low]
                series_corrected[is_too_high] = self.get_upper_limit(series=series)[is_too_high]
            elif self.outlier_correction_method == CorrectOutliersMethodNames.interpolate:
                series_corrected[is_too_low] = np.nan
                series_corrected[is_too_high] = np.nan
                series_corrected = series_corrected.interpolate()
            else:
                raise NotImplementedError
        return series_corrected.rename(series.name)

    def get_idx_too_high(self, series: pd.Series) -> pd.Series:
        return series > self.get_upper_limit(series=series)

    def get_idx_too_low(self, series: pd.Series) -> pd.Series:
        return series < self.get_lower_limit(series=series)

    def get_upper_limit(self, series: pd.Series) -> pd.Series:
        return self._get_smoothed_series(series=series) + self.six_sigma_multiplier * self._get_iqr_residuals(
            series=series
        )

    def get_lower_limit(self, series: pd.Series) -> pd.Series:
        return self._get_smoothed_series(series=series) - self.six_sigma_multiplier * self._get_iqr_residuals(
            series=series
        )

    def _get_iqr_residuals(self, series: pd.Series) -> float:
        residuals = series - self._get_smoothed_series(series=series)
        return residuals.quantile(q=0.75) - residuals.quantile(q=0.25)

    def _get_smoothed_series(self, series: pd.Series) -> pd.Series:
        return smooth_series(y=series, window=self.smoothing_window, quantile=0.5, ewm_alpha=self.ewm_alpha)



from __future__ import annotations

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin


class SeriesToFrameTransformer(BaseEstimator, TransformerMixin):
    def fit(self, X: pd.Series, y: pd.Series = None) -> SeriesToFrameTransformer:
        return self

    @staticmethod
    def transform(X: pd.Series) -> pd.DataFrame:
        return X.to_frame()



import numpy as np
import pandas as pd
import pytest
from pandas._testing import assert_series_equal

from hcl_model.transformers.outlier_correction import TargetOutlierCorrectionTransformer, CorrectOutliersMethodNames


class TestTargetOutlierCorrectionTransformer:
    def test_outliers(self) -> None:
        periods = 10
        index = pd.date_range(start="2021-01-01", periods=periods)
        series = pd.Series(0, index=index)
        too_low_idx = [1, 2, 4]
        too_high_idx = [6, 7, 9]
        series.iloc[too_high_idx] = 10
        series.iloc[too_low_idx] = -10

        outliers_detector = TargetOutlierCorrectionTransformer(
            outlier_correction_method=CorrectOutliersMethodNames.trim,
            six_sigma_multiplier=1e-10,
            smoothing_window=periods,
            ewm_alpha=1 / periods,
        )
        too_low_actual = outliers_detector.get_idx_too_low(series=series)
        too_low_expected = pd.Series(False, index=index)
        too_low_expected.iloc[too_low_idx] = True
        too_high_actual = outliers_detector.get_idx_too_high(series=series)
        too_high_expected = pd.Series(False, index=index)
        too_high_expected.iloc[too_high_idx] = True

        pd.testing.assert_series_equal(too_low_actual, too_low_expected)
        pd.testing.assert_series_equal(too_high_actual, too_high_expected)

        outliers_detector = TargetOutlierCorrectionTransformer(
            outlier_correction_method=CorrectOutliersMethodNames.trim,
            six_sigma_multiplier=1e-10,
            smoothing_window=1,
            ewm_alpha=1,
        )
        too_low_actual = outliers_detector.get_idx_too_low(series=series)
        too_low_expected = pd.Series(False, index=index)
        too_high_actual = outliers_detector.get_idx_too_high(series=series)
        too_high_expected = pd.Series(False, index=index)

        pd.testing.assert_series_equal(too_low_actual, too_low_expected)
        pd.testing.assert_series_equal(too_high_actual, too_high_expected)

        outliers_detector = TargetOutlierCorrectionTransformer(
            outlier_correction_method=CorrectOutliersMethodNames.trim,
            six_sigma_multiplier=1e10,
            smoothing_window=periods,
            ewm_alpha=1 / periods,
        )
        too_low_actual = outliers_detector.get_idx_too_low(series=series)
        too_low_expected = pd.Series(False, index=index)
        too_high_actual = outliers_detector.get_idx_too_high(series=series)
        too_high_expected = pd.Series(False, index=index)

        pd.testing.assert_series_equal(too_low_actual, too_low_expected)
        pd.testing.assert_series_equal(too_high_actual, too_high_expected)

    @pytest.mark.parametrize("y_type", ["series", "ndarray"])
    @pytest.mark.parametrize("six_sigma_multiplier", [3, 4, 5])
    def test_get_series_with_corrected_outliers_with_no_correction(
        self, y_type: str, six_sigma_multiplier: int
    ) -> None:
        date_col = "date"
        value_col = "value"
        size = 300
        values = np.zeros(size)
        values[10] = 40
        values[50] = 80
        values[150] = 60
        df = pd.DataFrame(
            {date_col: pd.date_range(start="2020-12-29", freq="W-SAT", periods=size), value_col: values}
        ).set_index(date_col)
        y = pd.Series(df[value_col], index=df.index, name=value_col)
        outliers_detector = TargetOutlierCorrectionTransformer(
            outlier_correction_method=CorrectOutliersMethodNames.nothing, six_sigma_multiplier=six_sigma_multiplier
        )
        y_corrected = outliers_detector.transform(X=y.values if y_type == "ndarray" else y)
        if y_type == "ndarray":
            np.array_equal(y.values, y_corrected)
        else:
            assert_series_equal(y, y_corrected)

    @pytest.mark.parametrize("y_type", ["series", "ndarray"])
    @pytest.mark.parametrize("six_sigma_multiplier", [3, 4, 5])
    @pytest.mark.parametrize(
        "outlier_correction_method", [CorrectOutliersMethodNames.trim, CorrectOutliersMethodNames.interpolate]
    )
    def test_get_series_with_corrected_outliers(
        self, y_type: str, six_sigma_multiplier: int, outlier_correction_method: CorrectOutliersMethodNames
    ) -> None:
        date_col = "date"
        value_col = "value"
        size = 300
        values = np.zeros(size)
        values[10] = 40
        values[50] = 80
        values[150] = 60
        df = pd.DataFrame(
            {date_col: pd.date_range(start="2020-12-29", freq="W-SAT", periods=size), value_col: values}
        ).set_index(date_col)
        y = pd.Series(df[value_col], index=df.index, name=value_col)
        outliers_detector = TargetOutlierCorrectionTransformer(
            outlier_correction_method=outlier_correction_method, six_sigma_multiplier=six_sigma_multiplier
        )
        y_corrected = outliers_detector.transform(X=y.values if y_type == "ndarray" else y)
        expected_nr_outliers_found = 3
        nr_outliers_found = sum(y != y_corrected)

        if y_type == "series":
            assert y.name == y_corrected.name
        assert sum(abs(y_corrected)) <= sum(abs(y))
        assert nr_outliers_found == expected_nr_outliers_found

    @pytest.mark.parametrize("y_type", ["series", "ndarray"])
    @pytest.mark.parametrize("six_sigma_multiplier", [3, 4, 5])
    @pytest.mark.parametrize(
        "outlier_correction_method", [CorrectOutliersMethodNames.trim, CorrectOutliersMethodNames.interpolate]
    )
    def test_get_series_with_corrected_outliers_constant_value_no_outliers(
        self, y_type: str, six_sigma_multiplier: int, outlier_correction_method: CorrectOutliersMethodNames
    ) -> None:
        date_col = "date"
        value_col = "value"
        size = 300
        values = np.zeros(size)
        df = pd.DataFrame(
            {date_col: pd.date_range(start="2020-12-29", freq="W-SAT", periods=size), value_col: values}
        ).set_index(date_col)
        y = pd.Series(df[value_col], index=df.index, name=value_col)
        outliers_detector = TargetOutlierCorrectionTransformer(
            outlier_correction_method=outlier_correction_method, six_sigma_multiplier=six_sigma_multiplier
        )
        y_corrected = outliers_detector.transform(X=y.values if y_type == "ndarray" else y)

        if y_type == "ndarray":
            np.array_equal(y.values, y_corrected)
        else:
            assert_series_equal(y, y_corrected)

    @pytest.mark.parametrize("y_type", ["series", "ndarray"])
    @pytest.mark.parametrize("six_sigma_multiplier", [3, 4, 5])
    @pytest.mark.parametrize(
        "outlier_correction_method", [CorrectOutliersMethodNames.trim, CorrectOutliersMethodNames.interpolate]
    )
    def test_get_series_with_corrected_outliers_no_outliers(
        self, y_type: str, six_sigma_multiplier: int, outlier_correction_method: CorrectOutliersMethodNames
    ) -> None:
        date_col = "date"
        value_col = "value"
        size = 300
        values = np.random.normal(size=size)
        df = pd.DataFrame(
            {date_col: pd.date_range(start="2020-12-29", freq="W-SAT", periods=size), value_col: values}
        ).set_index(date_col)
        y = pd.Series(df[value_col], index=df.index, name=value_col)
        outliers_detector = TargetOutlierCorrectionTransformer(
            outlier_correction_method=outlier_correction_method, six_sigma_multiplier=six_sigma_multiplier
        )
        y_corrected = outliers_detector.transform(X=y.values if y_type == "ndarray" else y)

        if y_type == "ndarray":
            np.array_equal(y.values, y_corrected)
        else:
            assert_series_equal(y, y_corrected)

    def test_get_series_with_outliers_and_with_wrong_method(self) -> None:
        date_col = "date"
        value_col = "value"
        size = 300
        values = np.random.normal(size=size)
        values[10] = 40
        values[50] = 80
        values[150] = 60
        df = pd.DataFrame(
            {date_col: pd.date_range(start="2020-12-29", freq="W-SAT", periods=size), value_col: values}
        ).set_index(date_col)
        y = pd.Series(df[value_col], index=df.index, name=value_col)

        with pytest.raises(NotImplementedError):
            outliers_detector = TargetOutlierCorrectionTransformer(
                outlier_correction_method="magic", six_sigma_multiplier=3
            )
            outliers_detector.transform(X=y)



import pandas as pd


def smooth_series(y: pd.Series, window: int, quantile: float, ewm_alpha: float) -> pd.Series:
    return y.rolling(window=window).quantile(quantile=quantile).fillna(method="bfill").ewm(alpha=ewm_alpha).mean()



from abc import ABC, abstractmethod
from typing import List, Optional, Union

import numpy as np
import pandas as pd

from hcl_model.utils.check_x_y import check_X_y


class ModelBase(ABC):
    """Time Series Model base class."""

    lbl_r2 = "rsquared"
    lbl_aic = "aic"
    lbl_mape = "mape"
    lbl_resid_mean = "error_mean"
    lbl_resid_std = "error_std"
    lbl_resid_skewness = "error_skewness"
    lbl_resid_kurtosis = "error_kurtosis"
    lbl_params = "parameters"

    def __init__(self) -> None:
        self._fit_results = None
        self._y_train = None
        self._x_train = None

    def fit(self, X: Optional[pd.DataFrame], y: Union[pd.Series, np.ndarray], **kwargs):
        """
        Fit the model using some provided training data.

        :param X: optional exogenous explanatory variables
        :param y: pd.Series or np.ndarray, endogenous variable
            In case of np.ndarray its index is inherited from `X`.
            In case of np.ndarray and if `X` is `None`, then `TypeError` is risen.
        """
        self._x_train, self._y_train = check_X_y(X=X, y=y)
        self._fit(**kwargs)
        return self

    @abstractmethod
    def _fit(self, **kwargs) -> None:
        pass

    def predict(
        self,
        X: Optional[pd.DataFrame],
        num_steps: int,
        quantile_levels: List[float] = None,
        num_simulations: int = None,
        **kwargs
    ) -> pd.DataFrame:
        """
        Forecast the values and prediction intervals

        :param X: exogenous variables should cover the whole prediction horizon
        :param num_steps: number of point in the future that we want to forecast
        :param quantile_levels: list of desired prediction interval levels between 0 and 100 (in percentages).
            If not provided, no confidence interval will be given as output
        :param num_simulations: number of simulations for simulation-based prediction intervals
        :returns: A DataFrame containing values and prediction intervals.

        Example of output from `num_steps=2` and `quantile_levels=[5, 95]`:

        .. code-block:: python

                        rate      q5     q95
            2019-06-07   102      75     127
            2019-06-14   305     206     278
        """
        self._check_exogenous(exog=X, nobs=self._nobs, num_steps=num_steps)
        predictions = self._predict(num_steps=num_steps, X=X, quantile_levels=quantile_levels, **kwargs)
        if quantile_levels is not None:
            quantiles = self._compute_prediction_quantiles(
                num_steps=num_steps, quantile_levels=quantile_levels, X=X, num_simulations=num_simulations
            )
            predictions = pd.concat([predictions, quantiles], axis=1)
        return self._add_trend(df=predictions)

    @abstractmethod
    def _predict(
        self, num_steps: int, X: pd.DataFrame = None, quantile_levels: List[float] = None, **kwargs
    ) -> pd.DataFrame:
        pass

    def simulate(self, num_steps: int, num_simulations: int, X: pd.DataFrame = None, **kwargs) -> pd.DataFrame:
        """
        Simulate `num_simulations` realizations of the next `num_steps` values

        :param num_steps: number of points in the future that we want to simulate
        :param num_simulations: number of independent simulations
        :param X: exogenous variables
        :return: A DataFrame containing simulations
        """
        self._check_exogenous(exog=X, nobs=self._nobs, num_steps=num_steps)
        return self._simulate(num_steps=num_steps, num_simulations=num_simulations, X=X, **kwargs)

    @abstractmethod
    def _simulate(self, num_steps: int, num_simulations: int, **kwargs) -> pd.DataFrame:
        pass

    @abstractmethod
    def _compute_prediction_quantiles(
        self, num_steps: int, num_simulations: int = None, quantile_levels: List[float] = None, X: pd.DataFrame = None
    ) -> pd.DataFrame:
        pass

    def summary(self) -> pd.Series:
        """A summary of in-sample model performance KPIs.

        :return: A series of model fit KPIs.
        """
        return pd.Series(
            {
                self.lbl_r2: self._get_rsquared(),
                self.lbl_aic: self._get_aic(),
                self.lbl_mape: self._get_mape(),
                self.lbl_resid_mean: self._get_residual_mean(),
                self.lbl_resid_std: self._get_residual_std(),
                self.lbl_resid_skewness: self._get_residual_skewness(),
                self.lbl_resid_kurtosis: self._get_residual_kurtosis(),
                self.lbl_params: self.get_parameters().to_dict(),
            }
        )

    @abstractmethod
    def _get_aic(self) -> float:
        """Akaike Information Criterion of a model fit.

        :return: AIC statistic as a float.
        """

    @abstractmethod
    def _get_fitted_values(self) -> pd.Series:
        """get fitted values

        :return: One point ahead forecasts on the in-sample period which are the "fitted values" in time series context.
        """

    @abstractmethod
    def _get_residuals(self) -> pd.Series:
        """Get residuals

        :return: Residuals of one point ahead forecasts on the in-sample period.
        """

    def _get_mape(self) -> float:
        """Mean absolute percentage error on in-sample.

        :return: Error as a float in percent.
        """
        return ((self._y_train - self._get_fitted_values()) / self._y_train).abs().mean() * 100

    def _get_rsquared(self) -> float:
        """Mean absolute percentage error on in-sample.

        :return: Error as a float in percent.
        """
        return np.corrcoef(self._get_fitted_values(), self._y_train.values)[0, 1] ** 2

    def _get_residual_moment(self, degree: int = 1, center_first: bool = False) -> float:
        """Get residual moment.

        :return: (Centered) moments of model fit residuals. Note that residuals can be biased on average.
        """
        if center_first:
            resid = self._get_residuals() - self._get_residuals().mean()
        else:
            resid = self._get_residuals()
        return resid.pow(degree).mean()

    def _get_residual_mean(self) -> float:
        return self._get_residual_moment(degree=1, center_first=False)

    def _get_residual_std(self) -> float:
        return self._get_residuals().std()

    def _get_residual_skewness(self) -> float:
        return self._get_residual_moment(degree=3, center_first=True) / (self._get_residual_std() ** 3)

    def _get_residual_kurtosis(self) -> float:
        return self._get_residual_moment(degree=4, center_first=True) / (self._get_residual_std() ** 4)

    def _check_exogenous(self, nobs: int, num_steps: int, exog: pd.DataFrame = None) -> None:
        """Check that provided exogenous data cover prediction horizon.

        :param nobs: the number of observations
        :param num_steps: the number of steps in the future we want to make forecast of
        :param exog: exogenous data
        :raise RuntimeError:
        """
        x_train_and_test = pd.concat([self._x_train, exog]) if exog is not None else self._x_train
        if x_train_and_test is not None and x_train_and_test.shape[0] < nobs + num_steps:
            raise RuntimeError("Provided exogenous data does not cover the whole prediction horizon!")

    def _get_endog_name(self) -> str:
        return self._y_train.name

    def _get_index_name(self) -> str:
        return self._y_train.index.name

    @property
    def _nobs(self) -> int:
        return self._y_train.shape[0]

    def get_parameters(self) -> pd.Series:
        """Get model parameters.

        :return: A series of model parameters.
        """
        return self._fit_results.params

    @staticmethod
    def get_quantile_names(quantile_levels: List[float]) -> List[str]:
        return ["prediction_quantile{}".format(x) for x in quantile_levels]

    @abstractmethod
    def _add_trend(self, df: pd.DataFrame) -> pd.DataFrame:
        pass



import pandas as pd

from hcl_model.utils.get_duplicate_columns import get_duplicate_columns


def test_get_duplicate_columns() -> None:
    students = [
        ("jack", 34, "Sydeny", 34, "Sydeny", 34),
        ("Riti", 30, "Delhi", 30, "Delhi", 30),
        ("Aadi", 16, "New York", 16, "New York", 16),
        ("Riti", 30, "Delhi", 30, "Delhi", 30),
        ("Riti", 30, "Delhi", 30, "Delhi", 30),
        ("Riti", 30, "Mumbai", 30, "Mumbai", 30),
        ("Aadi", 40, "London", 40, "London", 40),
        ("Sachin", 30, "Delhi", 30, "Delhi", 30),
    ]

    df = pd.DataFrame(students, columns=["Name", "Age", "City", "Marks", "Address", "Pin"])

    duplicate_column_names = get_duplicate_columns(df)

    assert set(duplicate_column_names) == {"Marks", "Address", "Pin"}



import numpy as np
import pandas as pd
from pandas._testing import assert_frame_equal

from hcl_model.transformers.series_to_frame import SeriesToFrameTransformer


def test_series_to_frame() -> None:
    df = SeriesToFrameTransformer().transform(X=pd.Series(dtype=float))
    assert_frame_equal(df, pd.DataFrame(columns=[0], dtype=np.float64))
    df = SeriesToFrameTransformer().transform(X=pd.Series(np.arange(5), name="value"))
    assert_frame_equal(df, pd.DataFrame({"value": np.arange(5)}))
    index = pd.date_range(start="2022-01-01", periods=5, name="date")
    df = SeriesToFrameTransformer().transform(X=pd.Series(np.arange(5), index=index, name="value"))
    assert_frame_equal(df, pd.DataFrame({"value": np.arange(5)}, index=index))



class LabelsCommon:
    date = "date"
    isoweek = "isoweek"
    month = "month"
    day = "day"


class LabelsExog:
    lbl_calendar_numeric = "calendar_numeric"
    lbl_spline_dim = "spline_dim"
    lbl_is_holiday = "is_holiday"


class LabelsHolidaysOut:
    lbl_country = "country"
    lbl_holiday = "holiday"



from typing import List, Union

import pandas as pd
from statsmodels.tsa.tsatools import add_trend

from hcl_model.transformers.calendar import CalendarTransformer
from hcl_model.utils.get_duplicate_columns import get_duplicate_columns


def construct_calendar_exogenous(
    endog: pd.Series,
    num_steps: int = 52,
    trend: str = "c",
    splines_df: int = None,
    holidays: List[dict] = None,
    auto_dummy_max_number: int = None,
    auto_dummy_threshold: float = 2,
) -> Union[None, pd.DataFrame]:
    """Construct deterministic exogenous variables.

    :param endog: time series of endogenous regressor
    :param num_steps: number of periods for forecasting.
        Output DataFrame will be longer than `endog` by this number of rows.
    :param trend: follows `statsmodels.tsa.tsatools.add_trend`
    :param splines_df: number of degrees of freedom for splines. 1 or more
    :param holidays: list of dicts with each dict representing one holiday.
        Each value of this series is a string representation of a dictionary.
        This dictionary is consumed by `utils.calendar_transformer.CalendarTransformer.add_holiday_dummies`.
    :param auto_dummy_max_number: limit on the number of automatic seasonal dummies
    :param auto_dummy_threshold: cutoff for "irregular" time series changes

    :return: DataFrame with exogenous regressors.
    """
    if endog.name is None:
        endog.name = "endog"

    extended = endog.reindex(
        pd.date_range(
            start=endog.index[0],
            periods=endog.shape[0] + num_steps,
            freq=pd.infer_freq(endog.index),
        )
    )

    df = add_trend(extended, trend=trend)
    cal_transformer = CalendarTransformer()

    if splines_df is not None:
        df = cal_transformer.add_periodic_splines(df, degrees_of_freedom=int(splines_df))

    if holidays is not None:
        for i, holiday in enumerate(holidays):
            if holiday is not None:
                df = cal_transformer.add_holiday_dummies(df, **holiday, dummy_name="holiday_{}".format(i + 1))

    if auto_dummy_max_number is not None:
        df = cal_transformer.add_automatic_seasonal_dummies(
            df=df,
            var_name=endog.name,
            threshold=auto_dummy_threshold,
            lim_num_dummies=auto_dummy_max_number,
        )

    return df.drop(columns=get_duplicate_columns(df)).iloc[:, 1:]



import random

import numpy as np
import pandas as pd
import pytest
from pandas._testing import assert_series_equal

from hcl_model.transformers.structural_breaks import TargetStructuralBreakCorrectionTransformer


@pytest.mark.parametrize("y_type", ["series", "ndarray"])
class TestStructuralBreakCorrectionTransformer:
    def test_no_correction_series_with_structural_break(self, y_type: str) -> None:
        series1 = np.ones(100)
        series2 = np.ones(100) + 1
        series = pd.Series(np.append(series1, series2))
        X = series.values if y_type == "ndarray" else series

        series_corrected = TargetStructuralBreakCorrectionTransformer(structural_break_correction=False).transform(X=X)

        assert len(series_corrected) == len(series)
        if y_type == "ndarray":
            np.array_equal(series.values, series_corrected)
        else:
            assert_series_equal(series, series_corrected)

    def test_correction_series_with_structural_break(self, y_type: str) -> None:
        series1 = np.ones(100)
        series2 = np.ones(100) + 1
        series = pd.Series(np.append(series1, series2))
        X = series.values if y_type == "ndarray" else series

        series_corrected = TargetStructuralBreakCorrectionTransformer().transform(X=X)
        series_corrected_expected = pd.Series(np.ones(200) + 1)

        assert len(series_corrected) == len(series)
        if y_type == "ndarray":
            np.array_equal(series_corrected_expected.values, series_corrected)
        else:
            assert_series_equal(series_corrected_expected, series_corrected)

    def test_adjusted_variability_still_zero(self, y_type: str) -> None:
        random.seed(101)
        series1 = pd.Series(np.ones(100))
        series2 = pd.Series(np.random.normal(size=100, loc=100))
        series = pd.concat([series1, series2])
        X = series.values if y_type == "ndarray" else series

        series_corrected = TargetStructuralBreakCorrectionTransformer().transform(X=X)
        series_corrected_first_part = series_corrected[:100]

        assert len(np.unique(series_corrected_first_part)) == len(np.unique(series1))

    def test_variability_adjusted(self, y_type: str) -> None:
        random.seed(101)
        series1 = pd.Series(np.random.normal(size=100, loc=1, scale=1))
        series2 = pd.Series(np.random.normal(size=100, loc=100, scale=5))
        series = pd.concat([series1, series2])
        X = series.values if y_type == "ndarray" else series

        series_corrected = TargetStructuralBreakCorrectionTransformer().transform(X=X)
        series_corrected_first_part = series_corrected[:100]

        assert series_corrected_first_part.std() > series1.std()

    def test_last_part_equal_to_original(self, y_type: str) -> None:
        random.seed(101)
        series1 = pd.Series(np.random.normal(size=100))
        series2 = pd.Series(np.random.normal(size=100, loc=100))
        series = pd.concat([series1, series2])
        X = series.values if y_type == "ndarray" else series

        series_corrected = TargetStructuralBreakCorrectionTransformer().transform(X=X)
        series_corrected_last_part = series_corrected[100:]

        if y_type == "ndarray":
            np.array_equal(series_corrected_last_part, series2)
        else:
            assert_series_equal(series_corrected_last_part, series2)

    def test_correction_series_without_structural_break(self, y_type: str) -> None:
        series = pd.Series(np.ones(200))
        X = series.values if y_type == "ndarray" else series

        series_corrected = TargetStructuralBreakCorrectionTransformer().transform(X=X)

        assert len(series_corrected) == len(series)
        if y_type == "ndarray":
            np.array_equal(series.values, series_corrected)
        else:
            assert_series_equal(series, series_corrected)



from __future__ import annotations

from typing import List, Union, Sequence, Dict, Callable, Optional

import numpy as np
import pandas as pd
from statsmodels.regression.linear_model import WLS

from hcl_model.model_base import ModelBase


class HandCraftedLinearModel(ModelBase):
    r"""Hand Crafted Linear Model

    :param endog_transform: transformation functions of the endogenous variable
    :param exog_transform: transformation functions of the exogenous variables

    Notes
    -----
    This class implements a general HCL model described `here <../hcl_model_math.html>`_.

    **Example of an endog_transform input:**

    .. code-block:: python

        f = {'lag1': lambda y: y.shift(1),
             'local_mean': lambda y: y.shift(1).ewm(span=5).mean()}

    Note that names for f must be chosen. These names will be used to call columns of the explanatory data.

    **Example of a exog_transform input:**

    Say, exog has column names 'x1', 'x2' and 'x3' but you only want to transform
    the first two. Then use

    .. code-block:: python

        g1 = lambda x: x**x
        g2 = lambda x: x+1
        g = {'x1': g1, 'x2': g2}

    This will result in 2 dimensional regressor matrix exog_transformed with just
    the transforms and without `'x3'`. If `g_transform=None`, then original exog is used.

    .. code-block:: python

        model = HandCraftedLinearModel(endog_transform=f, exog_transform=g)

    This function should run only if `endog_transform` is not `None`.

    **Example of weights:**

    .. code-block:: python

        weights = np.power(np.arange(start=0, stop=1, step=1/len(endog)), 0.5)

    Such a weighting scheme will put more focus on recent observations, with a rather slow decrease
    to 0 with increasing distance.

    """

    lbl_original_endog = "original_endog"
    lbl_original_exog = "original_exog"

    def __init__(self, endog_transform: Dict[str, Callable] = None, exog_transform: Dict[str, Callable] = None) -> None:
        super().__init__()
        self._endog_transform = self._init_transform(name=self.lbl_original_endog, transform=endog_transform)
        self._exog_transform = self._init_transform(name=self.lbl_original_exog, transform=exog_transform)

    def _fit(self, weights: Union[Sequence, float] = 1.0) -> None:
        transformed = self._transform_all_data(endog=self._y_train, exog=self._x_train)
        rhs_vars = self._convert_transformed_dict_to_frame(transformed=transformed)
        self._fit_results = WLS(endog=self._y_train, exog=rhs_vars, weights=weights, missing="drop").fit()

    def _predict(
        self,
        num_steps: int,
        X: pd.DataFrame = None,
        weights: Union[Sequence, float] = 1.0,
        quantile_levels: List[float] = None,
        num_simulations: int = None,
    ) -> pd.DataFrame:
        endog_updated = pd.concat(
            [self._y_train, pd.Series(np.empty(num_steps), name=self._get_endog_name(), index=X.index)]
        )
        x_train_and_test = pd.concat([self._x_train, X])
        for j in range(num_steps):
            transformed = self._transform_all_data(
                endog=endog_updated[: self._nobs + j + 1], exog=x_train_and_test.iloc[: self._nobs + j + 1]
            )
            rhs_vars = self._convert_transformed_dict_to_frame(transformed=transformed).iloc[-1, :]
            endog_updated.iloc[self._nobs + j] = np.dot(rhs_vars, self.get_parameters())

        return endog_updated.iloc[self._nobs :].to_frame().rename_axis(index=self._y_train.index.name)

    def _simulate(self, num_steps: int, num_simulations: int, X: pd.DataFrame = None, **kwargs) -> pd.DataFrame:
        num_params = self.get_parameters().shape[0]
        simulation = np.empty((num_steps, num_simulations))

        # simulate num_simulations parameters from its multivariate normal distribution
        # Cholesky decomposition returns only the lower triangular part,
        # so it has to be transposed before multiplication
        beta_simulated = (
            np.dot(
                np.random.normal(loc=0, scale=1, size=(num_simulations, num_params)),
                np.linalg.cholesky(self._fit_results.cov_params()).T,
            )
            + self.get_parameters().values
        )
        beta_simulated = pd.DataFrame(beta_simulated, columns=self._fit_results.params.index)
        # simulate innovations for the right hand side of the model
        innovation = np.random.normal(
            loc=0, scale=self._fit_results.mse_resid ** 0.5, size=(num_steps, num_simulations)
        )

        # stack observed endogenous series and empty container for future simulations
        # Series of length (nobs + num_steps)
        endog_updated = pd.concat([self._y_train, pd.Series(np.empty(num_steps))], axis=0, ignore_index=True)
        # DataFrame (nobs + num_steps) x num_simulations
        endog_updated = pd.concat([endog_updated] * num_simulations, axis=1)

        # loop over horizon
        for j in range(num_steps):
            transformed_endog = self._transform_all_data(endog=endog_updated.iloc[: self._nobs + j + 1])
            # apply model recursion plus innovation
            temp = 0
            for key, val in transformed_endog.items():
                temp += val.iloc[-1, :] * beta_simulated[key]

            transformed_exog = self._transform_all_data(exog=pd.concat([self._x_train, X]).iloc[: self._nobs + j + 1])
            exog_df = self._convert_transformed_dict_to_frame(transformed=transformed_exog)
            for key in exog_df.columns:
                temp += exog_df[key].iloc[-1] * beta_simulated[key]

            simulation[j] = temp + innovation[j]
            # update out-of-sample endogenous series with simulated value
            endog_updated.loc[self._nobs + j] = simulation[j]

        return pd.DataFrame(simulation, index=X.index)

    def _compute_prediction_quantiles(
        self, num_steps: int, num_simulations: int = None, quantile_levels: List[float] = None, X: pd.DataFrame = None
    ) -> pd.DataFrame:
        simulations = self.simulate(num_steps=num_steps, num_simulations=num_simulations, X=X)
        quantiles = simulations.quantile(np.array(quantile_levels) / 100, axis=1).T
        quantiles.columns = self.get_quantile_names(quantile_levels)
        return quantiles

    def _get_rsquared(self) -> float:
        return self._fit_results.rsquared

    def _get_aic(self) -> float:
        return self._fit_results.aic

    def _get_fitted_values(self) -> pd.Series:
        return self._fit_results.fittedvalues

    def _get_residuals(self) -> pd.Series:
        return self._fit_results.resid

    @staticmethod
    def _transform_data(
        transform: Dict[str, Callable], data: Union[pd.Series, pd.DataFrame] = None
    ) -> Dict[str, Union[pd.Series, pd.DataFrame]]:
        """Transform original data for the use as right-hand-side variables.

        :param data: original data
        :param transform: dictionary with transformation functions
        :return: dictionary with the same keys as in transform dictionary
        """
        return {key: fun(data) for key, fun in transform.items() if data is not None}

    def _transform_all_data(
        self, endog: Union[pd.Series, pd.DataFrame] = None, exog: pd.DataFrame = None
    ) -> Dict[str, Union[pd.Series, pd.DataFrame]]:
        """Transform data according to the model.

        :param endog: endogenous time series or frame (when doing Monte Carlo)
        :param exog: frame of exogenous variables
        :return: dictionary {f(Y_{t,L}), g(X_t)} of transformed endogenous and exogenous.
        Original endogenous variable is dropped from the dictionary to avoid using it as a right-hand-side variable.
        """
        transformed_endog = self._transform_data(data=endog, transform=self._endog_transform)
        transformed_exog = self._transform_data(data=exog, transform=self._exog_transform)
        transformed = {**transformed_endog, **transformed_exog}
        transformed.pop(self.lbl_original_endog, None)
        return transformed

    @staticmethod
    def _convert_transformed_dict_to_frame(transformed: Dict[str, Union[pd.Series, pd.DataFrame]]) -> pd.DataFrame:
        """Convert the dictionary of data into one DataFrame.

        :param transformed: dictionary of transformed data
        :return: DataFrame with transformed data.
        Column names in the resulting frame will be concatenated from the dictionary keys
        and column names of the inputs.
        """
        out = list()
        for key, frame in transformed.items():
            if isinstance(frame, pd.Series):
                out.append(pd.DataFrame({key: frame}))
            elif isinstance(frame, pd.DataFrame):
                out.append(frame.rename(columns={col: "{} {}".format(key, col) for col in frame.columns}))
        if len(out) > 0:
            return pd.concat(out, axis=1)
        else:
            return pd.DataFrame()

    @staticmethod
    def _init_transform(name: str, transform: Dict[str, Callable] = None) -> Optional[Dict[str, Callable]]:
        return {name: lambda x: x} if transform is None else transform

    def _add_trend(self, df: pd.DataFrame) -> pd.DataFrame:
        return df



from abc import abstractmethod
from typing import Tuple

import pandas as pd
import numpy as np
import pytest
from statsmodels.tsa.tsatools import add_trend

from hcl_model.model_hcl_generic import HandCraftedLinearModel
from hcl_model.model_sarimax import SARIMAXModel


class TestModelCommon:
    lbl_date = "date"
    lbl_value = "value"

    def generate_data(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        nobs = 30
        index = pd.date_range("2019-01-01", periods=nobs, freq="W-FRI", name=self.lbl_date)
        endog = pd.DataFrame({self.lbl_value: np.arange(1, nobs + 1) + np.random.normal(size=nobs)}, index=index)
        exog = pd.DataFrame({"const": np.ones(nobs), "time": np.arange(1, nobs + 1)}, index=index)
        return endog, exog

    @abstractmethod
    def test_model_fit(self):
        pass

    @abstractmethod
    def test_model_prediction(self):
        pass

    @abstractmethod
    def test_model_simulation(self):
        pass

    @abstractmethod
    def test_model_percentiles(self):
        pass

    @abstractmethod
    def test_model_summary(self):
        pass


@pytest.mark.parametrize("y_type", ["series", "ndarray"])
class TestPredictionsSanity:
    lbl_value = "value"

    def test_constant(self, y_type: str):
        nobs = 30
        num_steps = 10

        index = pd.date_range("2019-01-01", periods=nobs, freq="W-FRI", name="date")
        data = pd.Series(np.arange(1, nobs + 1), index=index, name=self.lbl_value)
        exog = add_trend(data, trend="c", has_constant="add").drop(self.lbl_value, axis=1)
        y_train = data.iloc[:-num_steps]
        x_train = exog.iloc[:-num_steps]
        x_test = exog.iloc[-num_steps:]

        expected_forecast = pd.DataFrame({self.lbl_value: y_train.mean()}, index=index[-num_steps:])

        model = SARIMAXModel(trend="c")
        model.fit(y=y_train, X=None)
        forecast_sarimax = model.predict(num_steps=num_steps, X=None)

        model = HandCraftedLinearModel()
        model.fit(y=y_train.values if y_type == "ndarray" else y_train, X=x_train)
        forecast_hcl = model.predict(num_steps=num_steps, X=x_test)

        pd.testing.assert_frame_equal(forecast_sarimax, expected_forecast)
        pd.testing.assert_frame_equal(forecast_hcl, expected_forecast)

    def test_linear_trend(self, y_type: str):
        nobs = 30
        num_steps = 10
        index = pd.date_range("2019-01-01", periods=nobs, freq="W-FRI", name="date")
        data = pd.Series(np.arange(1, nobs + 1), index=index, name=self.lbl_value, dtype=float)
        exog = add_trend(data, trend="ct", has_constant="add").drop(self.lbl_value, axis=1)
        y_train = data.iloc[:-num_steps]
        x_train = exog.iloc[:-num_steps]
        x_test = exog.iloc[-num_steps:]

        expected_forecast = pd.DataFrame({self.lbl_value: data.iloc[-num_steps:].values}, index=index[-num_steps:])

        model = SARIMAXModel(trend="ct")
        model.fit(y=y_train, X=None)
        forecast_sarimax = model.predict(num_steps=num_steps, X=None)

        model = HandCraftedLinearModel()
        model.fit(y=y_train.values if y_type == "ndarray" else y_train, X=x_train)
        forecast_hcl = model.predict(num_steps=num_steps, X=x_test)

        pd.testing.assert_frame_equal(forecast_sarimax, expected_forecast)
        pd.testing.assert_frame_equal(forecast_hcl, expected_forecast)

    def test_ar1_with_const(self, y_type: str):
        nobs = 30
        num_steps = 10
        index = pd.date_range("2019-01-01", periods=nobs, freq="W-FRI", name="date")
        data = pd.Series(np.arange(1, nobs + 1), index=index, name=self.lbl_value, dtype=float)
        exog = add_trend(data, trend="c", has_constant="add").drop(self.lbl_value, axis=1)
        y_train = data.iloc[:-num_steps]
        x_train = exog.iloc[:-num_steps]
        x_test = exog.iloc[-num_steps:]
        endog_transform = {"lag1": lambda y: y.shift(1)}

        model = SARIMAXModel(trend="n", order=(1, 0, 0), enforce_stationarity=False)
        model.fit(y=y_train.values if y_type == "ndarray" else y_train, X=x_train)
        forecast_sarimax = model.predict(num_steps=num_steps, X=x_test)

        model = HandCraftedLinearModel(endog_transform=endog_transform)
        model.fit(y=y_train.values if y_type == "ndarray" else y_train, X=x_train)
        forecast_hcl = model.predict(num_steps=num_steps, X=x_test)

        pd.testing.assert_frame_equal(forecast_sarimax, forecast_hcl, rtol=1e-1)

    def test_ar1_with_linear_trend(self, y_type: str):
        nobs = 30
        num_steps = 10
        index = pd.date_range("2019-01-01", periods=nobs, freq="W-FRI", name="date")
        data = pd.Series(np.arange(1, nobs + 1), index=index, name=self.lbl_value, dtype=float)
        exog = add_trend(data, trend="ct", has_constant="add").drop(self.lbl_value, axis=1)
        y_train = data.iloc[:-num_steps]
        x_train = exog.iloc[:-num_steps]
        x_test = exog.iloc[-num_steps:]
        endog_transform = {"lag1": lambda y: y.shift(1)}

        model = SARIMAXModel(trend="n", order=(1, 0, 0))
        model.fit(y=y_train.values if y_type == "ndarray" else y_train, X=x_train)
        forecast_sarimax = model.predict(num_steps=num_steps, X=x_test)

        model = HandCraftedLinearModel(endog_transform=endog_transform)
        model.fit(y=y_train.values if y_type == "ndarray" else y_train, X=x_train)
        forecast_hcl = model.predict(num_steps=num_steps, X=x_test)

        pd.testing.assert_frame_equal(forecast_sarimax, forecast_hcl, rtol=1e-1)



from __future__ import annotations

from enum import Enum
from typing import List, ValuesView


class StringEnum(str, Enum):
    @classmethod
    def get_all_values(cls) -> List[str]:
        return [v.value for v in cls._get_all_values()]

    @classmethod
    def _get_all_members(cls) -> List:
        return list(cls._get_all_values())

    @classmethod
    def _get_all_values(cls) -> ValuesView:
        return cls.__members__.values()



from typing import Iterable

import pandas as pd
from patsy import dmatrix
from scipy.stats import median_abs_deviation

from hcl_model.labels import LabelsCommon, LabelsExog
from hcl_model.utils.calendar_reader import CalendarReader


class CalendarTransformer:
    """
    Allows to easily add a variety of calendar-based columns to the exogenous regressor matrix.
    """

    lbl = LabelsCommon()
    lbl_exog = LabelsExog()

    lbl_weekly_freq = "7 D"
    lbl_auto_dummy = "automatic_dummy_{}"

    def __init__(self):
        self._cal_reader = CalendarReader()

    def add_periodic_splines(
        self,
        df: pd.DataFrame,
        input_level: str = "isoweek",
        degrees_of_freedom: int = 4,
    ) -> pd.DataFrame:
        """
        Add periodic spline scores to the dataframe df.

        df must have a datetime column as index. This method adds periodic splines
        scores to the dataframe by using its index as input.

        Typical output:

        .. code-block:: python

            df = pd.DataFrame({'x1': np.random.normal(0, 1, 53 * 3)},
                               index=pd.date_range(start='2015-01-01', freq='7 D', periods=53 * 3))

            cal = CalendarTransformer()
            cal.add_periodic_splines(df).head()

                              x1  spline_dim_1  spline_dim_2  spline_dim_3  spline_dim_4
            2015-01-01 -0.301179     -0.439735     -0.438071     -0.438801     -0.437531
            2015-01-08  0.301794     -0.343647     -0.460494     -0.409252     -0.498250
            2015-01-15 -0.431848     -0.218113     -0.474556     -0.374233     -0.526180
            2015-01-22  2.669445     -0.071386     -0.479289     -0.335690     -0.526172
            2015-01-29  0.220345      0.088282     -0.473725     -0.295568     -0.503074

        """
        data = df.copy(deep=True)
        data[self.lbl_exog.lbl_calendar_numeric] = self._get_numeric_periodicity(
            data.index, input_level=input_level
        ).astype(float)

        # the 'cc' part in the string of the dmatrix call is responsible for periodic splines.
        formula = f"cc({self.lbl_exog.lbl_calendar_numeric}, df={degrees_of_freedom}, constraints='center') - 1"
        calendar_matrix = dmatrix(formula, data=data, return_type="dataframe")

        # rename columns for convenience
        cols = calendar_matrix.columns
        calendar_matrix = calendar_matrix.rename(
            columns={col: f"{self.lbl_exog.lbl_spline_dim}_{num + 1}" for num, col in enumerate(cols)}
        )
        data[calendar_matrix.columns] = calendar_matrix

        return data.drop(self.lbl_exog.lbl_calendar_numeric, axis=1)

    def _get_numeric_periodicity(self, dates: pd.DatetimeIndex, input_level: str = "isoweek") -> pd.Series:
        """
        Add a column of numeric scores between 0 and 1 that represent annual periodicity.
        Input levels 'isoweek' and 'month' supported currently. For any other input we fall
        back to day of the year. The effect of using week or month is that all dates within one
        week or month get the same score while fall-back routine assigns each day of year same score.

        dates must have datetime data type.
        """
        if input_level == self.lbl.isoweek:
            return dates.isocalendar().week / 53
        elif input_level == self.lbl.month:
            return dates.month / 12
        elif input_level == self.lbl.day:
            return dates.isocalendar().day / 366
        else:
            raise ValueError(f"Unknown input_level argument: {input_level}")

    def add_holiday_triangles(
        self,
        df: pd.DataFrame,
        holiday_name: str,
        country_code: str,
        triangle_name: str = "",
        lh_factors: Iterable[float] = (1.0, 2.0, 3.0),
        rh_factors: Iterable[float] = (2.0, 1.0),
        by: str = None,
        merge_factors: bool = False,
    ) -> pd.DataFrame:
        """
        Adds triangular calendar regressors to the dataframe df. Index of df must be sorted and of date type.
        If there are additional grouping columns, then this can be set with the `by` argument (see pd.merge_asof).

        The user must pass two list, specifying the factors values that will be used.
        lh_factors = [..., factor one week before the holiday, factor on the week of the holiday]
        rh_factors = [factor on the week of the holiday, factor one week after the holiday, ...]

        If `merge_factors` = False, then a single numeric column of holiday scores is added. If not, two columns
        are added, representing effects before (including at) and after the holiday, respectively.

        Typical output:

        .. code-block:: python

            df = pd.DataFrame({'x1': np.random.normal(0, 1, 53 * 3)},
                               index=pd.date_range(start='2015-01-01', freq='7 D', periods=53 * 3))

            cal = CalendarTransformer()
            cal.add_holiday_triangles(df, 'New year', 'DE').head()

                              x1  _lh_factor  _rh_factor
            2015-01-01 -0.301179         3.0         2.0
            2015-01-08  0.301794         0.0         1.0
            2015-01-15 -0.431848         0.0         0.0
            2015-01-22  2.669445         0.0         0.0
            2015-01-29  0.220345         0.0         0.0
        """
        lbl_lh = "{}_lh_factor".format(triangle_name)
        lbl_rh = "{}_rh_factor".format(triangle_name)

        dates = self._read_holiday_dates(holiday_name, country_code)

        lh_df = list()
        for i, factor in enumerate(reversed([0.0] + list(lh_factors))):

            if merge_factors and i == 0:
                # if the user required the lh and rh factors to be merged in one, take only half value for the
                # day shared between lh and rh
                factor = factor * 0.5

            lh_df.append(
                pd.DataFrame(
                    {
                        self.lbl.date: dates[self.lbl.date] - pd.Timedelta("{} D".format(7 * i)),
                        lbl_lh: [factor] * len(dates),
                    }
                )
            )
        lh_df = pd.concat(lh_df).sort_values([self.lbl.date], ascending=True)

        # - build right-hand holiday factors
        rh_df = list()
        for i, factor in enumerate(list(rh_factors) + [0.0]):

            if merge_factors and i == 0:
                # if the user required the lh and rh factors to be merged in one, take only half value for the
                # day shared between lh and rh
                factor = factor * 0.5

            rh_df.append(
                pd.DataFrame(
                    {
                        self.lbl.date: dates[self.lbl.date] + pd.Timedelta("{} D".format(7 * i)),
                        lbl_rh: [factor] * len(dates),
                    }
                )
            )
        rh_df = pd.concat(rh_df).sort_values([self.lbl.date], ascending=True)

        # - generate new DataFrame with all holidays factors
        edf = pd.merge(left=lh_df, right=rh_df, how="outer", on=self.lbl.date).fillna(0).reset_index(drop=True)

        # - if the user required the lh and rh factors to be merged in one, merge them by summing the two columns
        # NOTE: the lh and rh factors in the case of merged factor are built to make the sum of the two columns
        # producing the merged factor
        if merge_factors:
            edf["{}_factor".format(triangle_name)] = edf[lbl_lh] + edf[lbl_rh]
            edf = edf.drop(columns=[lbl_lh, lbl_rh])

        # - insert all other missing days in the calendar, with factor 0
        edf = (
            edf.set_index(self.lbl.date)
            .sort_index()
            .reindex(
                pd.date_range(
                    start="{}-01-01".format(dates[self.lbl.date].dt.year.min()),
                    end="{}-12-31".format(dates[self.lbl.date].dt.year.max()),
                    freq="D",
                ),
                method="nearest",
            )
            .reset_index()
            .rename(columns={"index": self.lbl.date})
        )

        # - remove the time from the date column
        edf[self.lbl.date] = pd.to_datetime(edf[self.lbl.date].dt.date)

        return pd.merge_asof(
            df,
            edf.set_index(self.lbl.date),
            left_index=True,
            right_index=True,
            by=by,
            tolerance=pd.Timedelta(self.lbl_weekly_freq),
        )

    def add_holiday_dummies(
        self,
        df: pd.DataFrame,
        holiday_name: str,
        country_code: str,
        dummy_name: str = None,
        add_td: str = None,
        lags: Iterable[int] = (-2, -1, 0, 1, 2),
        by: str = None,
    ) -> pd.DataFrame:
        """
        Adds calendar effects to dataframe df with either 0 or 1 in each entry. 1 indicates that the corresponding
        date is close to the holiday that is considered.

        Inputs:
            df: dataframe with datetime index.
            holiday_name: name of the holiday. See CalendarReader.
            country_code: calendar's country code. See CalendarReader.
            dummy_name: optional name for renaming the dummy columns.
            add_td: adds to each holiday a fixed time delta. See add_cc_dummies as a typical example.
            lags: list of integers; add forward and backward looking lags of the dummy.
            by: name of group column if df should be grouped by that column when the merge_asof is applied.
        Output:
            returns df extended by dummy columns.

        Typical output:

        .. code-block:: python

            df = pd.DataFrame({'x1': np.random.normal(0, 1, 53 * 3)},
                               index=pd.date_range(start='2015-01-01', freq='7 D', periods=53 * 3))

            cal = CalendarTransformer()
            cal.add_holiday_dummies(df, 'Chinese New Year', 'TW').iloc[7:12, :]

                              x1  is_holiday  ...  is_holiday_lag_1  is_holiday_lag_2
            2015-02-19 -1.127138         0.0  ...               0.0               0.0
            2015-02-26  0.303294         1.0  ...               0.0               0.0
            2015-03-05 -0.398241         0.0  ...               1.0               0.0
            2015-03-12  1.281731         0.0  ...               0.0               1.0
            2015-03-19  2.303174         0.0  ...               0.0               0.0
        """
        dummy = self._get_holiday_dummy(holiday_name, country_code, add_td)

        if dummy_name is not None:
            dummy = dummy.rename(columns={self.lbl_exog.lbl_is_holiday: dummy_name})

        return self._add_dummy_lags_and_merge(df, dummy, lags, by)

    def _get_holiday_dummy(self, holiday_name: str, country_code: str, add_td: str = None) -> pd.DataFrame:
        """
        Any holiday included in the input dataframe can be translated to dummy column.
        Use e.g. `holiday_name` = 'Chinese New Year' and `country_code` = 'CN'.
        """
        dates = self._read_holiday_dates(holiday_name, country_code).assign(**{self.lbl_exog.lbl_is_holiday: 1})
        if add_td is not None:
            dates[self.lbl.date] = dates[self.lbl.date] + pd.Timedelta(add_td)

        # TODO: this is a very dirty quick fix!!!
        dates[self.lbl.date] -= pd.Timedelta("7 D")

        return pd.merge_asof(
            pd.DataFrame({self.lbl.date: self._get_weekly_date_range()}),
            dates,
            on=self.lbl.date,
            allow_exact_matches=False,
            tolerance=pd.Timedelta(self.lbl_weekly_freq),
        ).fillna(0)

    def _add_dummy_lags_and_merge(self, df: pd.DataFrame, dummy, lags: Iterable[int], by: str) -> pd.DataFrame:
        """
        Add optional forward and backward looking lags and merge to dataframe.
        """
        dummy_name = dummy.columns[-1]
        for lag in lags:
            lag_name = "{}_lag_{}".format(dummy_name, str(lag).replace("-", "neg"))
            dummy[lag_name] = dummy[dummy_name].shift(lag)

        # Add 0 to lags to preserve original dummy and rename:
        dummy = dummy.drop(dummy_name, axis=1).rename(columns={"{}_lag_0".format(dummy_name): dummy_name})

        return pd.merge_asof(
            df,
            dummy.set_index(self.lbl.date),
            left_index=True,
            right_index=True,
            by=by,
            tolerance=pd.Timedelta(self.lbl_weekly_freq),
        )

    def _get_weekly_date_range(self, start: str = "2005-01-03", periods: int = 53 * 25) -> pd.DatetimeIndex:
        return pd.date_range(start=start, periods=periods, freq=self.lbl_weekly_freq)

    def _read_holiday_dates(
        self,
        holiday_name: str,
        country_code: str,
        from_year: int = 2000,
        to_year: int = 2030,
    ) -> pd.DataFrame:
        return (
            self._cal_reader.get_holidays(holiday_name, country_code, from_year, to_year)
            .reset_index()
            .filter(items=[self.lbl.date])
        )

    @classmethod
    def add_automatic_seasonal_dummies(
        cls,
        df: pd.DataFrame,
        var_name: str,
        lim_num_dummies: int = 5,
        threshold: float = 3,
    ) -> pd.DataFrame:
        """Add automatic seasonal dummies.

        Outliers among weekly percentage changes are detected by normalizing and comparing with a certain threshold.
        Only weekly frequency is supported.

        :param df: original data
        :param var_name: the name of the variable to model and forecast
        :param lim_num_dummies: limit on the number of seasonal dummies
        :param threshold: quantile cutoff for "irregular" time series changes
        :return: original data with new columns corresponding to seasonal dummies
        """
        freq = pd.infer_freq(df.index)
        if freq[0] != "W":
            raise RuntimeError("Only weekly data is supported. Frequency detected: {}".format(freq))

        lbl_diff = "diff"
        lbl_week_number = "week_number"
        data = df.copy()
        data[lbl_week_number] = data.index.map(lambda x: x.isocalendar()[1])
        data[lbl_diff] = data[var_name] - data[var_name].ewm(com=10).mean()
        data[lbl_diff] = (data[lbl_diff] / data[lbl_diff].std()).abs()
        mean_abs_diff = data.iloc[10:].groupby(lbl_week_number)[lbl_diff].mean().dropna()
        normalized = (
            (mean_abs_diff - mean_abs_diff.median()).abs() / median_abs_deviation(mean_abs_diff, scale=1 / 1.4826 ** 2)
        ).sort_values(ascending=False)
        weeks = normalized.loc[normalized > threshold].index[:lim_num_dummies]
        for week in weeks:
            data[cls.lbl_auto_dummy.format(week)] = 0.0
            data.loc[
                data.index.map(lambda x: x.isocalendar()[1]) == week,
                cls.lbl_auto_dummy.format(week),
            ] = 1
        return data.drop([lbl_week_number, lbl_diff], axis=1)



import pandas as pd
import pytest

from hcl_model.transformers.truncate import TruncateTransformer


@pytest.mark.parametrize("full_weight_weeks", [1, 2, 31, 32])
def test_truncate(full_weight_weeks: int) -> None:
    lbl_date = "date"
    lbl_value = "value"
    df = pd.DataFrame({lbl_date: pd.date_range(start="2021-01-01", end="2021-02-01"), lbl_value: range(32)})
    s = df.set_index(lbl_date)[lbl_value]

    s_actual = TruncateTransformer(full_weight_weeks=full_weight_weeks).transform(X=s)
    s_expected = s[s.index > s.index.max() - pd.DateOffset(weeks=full_weight_weeks, days=1)]
    pd.testing.assert_series_equal(s_actual, s_expected)

    s_actual = TruncateTransformer(full_weight_weeks=len(s) + 1).transform(X=s)
    pd.testing.assert_series_equal(s_actual, s)

    s_actual = TruncateTransformer(full_weight_weeks=full_weight_weeks, truncate=False).transform(X=s)
    pd.testing.assert_series_equal(s_actual, s)

    data = df.set_index(lbl_date)
    df_actual = TruncateTransformer(full_weight_weeks=full_weight_weeks).transform(X=data)
    df_expected = data[data.index > data.index.max() - pd.DateOffset(weeks=full_weight_weeks, days=1)]
    pd.testing.assert_frame_equal(df_actual, df_expected)

    df_actual = TruncateTransformer(full_weight_weeks=data.shape[0] + 1).transform(X=data)
    pd.testing.assert_frame_equal(df_actual, data)

    df_actual = TruncateTransformer(full_weight_weeks=full_weight_weeks, truncate=False).transform(X=data)
    pd.testing.assert_frame_equal(df_actual, data)



from __future__ import annotations

from typing import Union

import pandas as pd
import pytest
from pandas._testing import assert_frame_equal
from sklearn.base import BaseEstimator, TransformerMixin

from hcl_model.transformers.feature_union import make_union_of_datetime_indexed_features


class TransformerForTests(BaseEstimator, TransformerMixin):
    def __init__(self, dates: Union[list[int], pd.DatetimeIndex], cols: Union[str, list[str]]) -> None:
        self._dates = dates
        self._cols = cols

    def fit(self, X: pd.DataFrame, y: pd.Series = None) -> TransformerForTests:
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        return X.loc[self._dates, self._cols]


def test_datetime_indexed_feature_union() -> None:
    date_index = pd.date_range(start="2021-01-01", periods=4)
    df = pd.DataFrame({"a": [1, 3, 6, 9], "b": [2, 5, 7, 10]}, index=date_index)
    feature_transformers = [
        TransformerForTests(dates=date_index, cols=["a"]),
        TransformerForTests(dates=date_index, cols=["b"]),
    ]
    feature_union = make_union_of_datetime_indexed_features(*feature_transformers)
    assert_frame_equal(feature_union.transform(X=df), df)
    assert_frame_equal(feature_union.fit_transform(X=df), df)

    feature_transformers = [
        TransformerForTests(dates=date_index[[0, 2]], cols=["a"]),
        TransformerForTests(dates=date_index[[1, 3]], cols=["b"]),
    ]
    feature_union = make_union_of_datetime_indexed_features(*feature_transformers)
    assert_frame_equal(feature_union.transform(X=df), df.loc[[]].set_index(pd.Index([], dtype="object")))
    assert_frame_equal(feature_union.fit_transform(X=df), df.loc[[]].set_index(pd.Index([], dtype="object")))

    feature_transformers = [
        TransformerForTests(dates=date_index[[0, 2]], cols=["a"]),
        TransformerForTests(dates=date_index, cols=["b"]),
    ]
    feature_union = make_union_of_datetime_indexed_features(*feature_transformers)
    assert_frame_equal(feature_union.transform(X=df), df.loc[date_index[[0, 2]]])
    assert_frame_equal(feature_union.fit_transform(X=df), df.loc[date_index[[0, 2]]])

    feature_transformers = [
        TransformerForTests(dates=date_index[[0, 2]], cols="a"),
        TransformerForTests(dates=date_index, cols=["b"]),
    ]
    feature_union = make_union_of_datetime_indexed_features(*feature_transformers)
    assert_frame_equal(feature_union.transform(X=df), df.loc[date_index[[0, 2]]])
    assert_frame_equal(feature_union.fit_transform(X=df), df.loc[date_index[[0, 2]]])

    feature_transformers = [
        TransformerForTests(dates=date_index[[0, 2]], cols="a"),
        TransformerForTests(dates=date_index, cols="b"),
    ]
    feature_union = make_union_of_datetime_indexed_features(*feature_transformers)
    assert_frame_equal(feature_union.transform(X=df), df.loc[date_index[[0, 2]]])
    assert_frame_equal(feature_union.fit_transform(X=df), df.loc[date_index[[0, 2]]])

    feature_transformers = [
        TransformerForTests(dates=[0, 2], cols="a"),
        TransformerForTests(dates=[0, 1, 2, 3], cols="b"),
    ]
    feature_union = make_union_of_datetime_indexed_features(*feature_transformers)
    error_msg = "All features should be in datetime-indexed Series or DataFrames"
    with pytest.raises(ValueError, match=error_msg):
        feature_union.transform(X=df.reset_index(drop=True))
    with pytest.raises(ValueError, match=error_msg):
        feature_union.fit_transform(X=df.reset_index(drop=True))
